import tifffile as tfi
import numpy as np
from skimage.filters import threshold_local
from scipy.ndimage.morphology import binary_opening,binary_erosion
import skimage.morphology as sm
from skimage.segmentation import watershed
from skimage import io, exposure, data
from skimage import measure
import os
import pandas as pd
from scipy.ndimage.morphology import binary_fill_holes
import skimage
from skimage.transform import rescale, resize, downscale_local_mean
from PIL import Image, ImageEnhance, ImageDraw,ImageFont
from AIPyS.supportFunctions.display_and_xml import evaluate_image_output,test_image



class Compsite_display:
    def __init__(self,input_image, mask_roi,channel = None):
        self.input_image = input_image
        self.mask_roi = mask_roi
        self.channel = channel
        self.rgb_out = self.draw_ROI_contour()

    def outline_seg(mask,index):
        '''
        outlined single ROI
        mask: binary string, "no-mask" or "mask" (inherit mask from __init__)
        index = ROI value
        :return
        '''
        seg_mask_temp = np.zeros(np.shape(mask), dtype=np.int32)
        seg_mask_temp[mask == index] = index
        seg_mask_eros_9 = binary_erosion(seg_mask_temp, structure=np.ones((9, 9))).astype(np.float64)
        seg_mask_eros_3 = binary_erosion(seg_mask_temp, structure=np.ones((3, 3))).astype(np.float64)
        framed_mask = seg_mask_eros_3 - seg_mask_eros_9
        return framed_mask

    def binary_frame_mask(ch,mask, table=None):
        '''
        Create a mask for NIS-elements to photo-activate for multiple point
        :parameter
        input_image_str: binary string,input image is-  "seed" or "target" (inherit from self)
        mask_str - input mask (RGB) 32integer
        :return
        framed_mask (RGB)
        '''
        if table is None:
            info_table = pd.DataFrame(
                measure.regionprops_table(
                    mask,
                    intensity_image=ch,
                    properties=['area', 'label', 'centroid'],
                )).set_index('label')
        else:
            info_table = table
        info_table['label'] = range(2, len(info_table) + 2)
        framed_mask = np.zeros(np.shape(mask), dtype=np.int32)
        if len(info_table['label']) > 1:
            for i in info_table.index.values:
                framed_mask = framed_mask + Compsite_display.outline_seg(mask, i)
        else:
            framed_mask = Compsite_display.outline_seg(mask, info_table.index.values)
        return framed_mask

    def draw_ROI_contour(self):
        ''' create a 3 channel rgb image from 16bit input image
            optional bin contour image from ROI image
            :parameter
            input_image: binary string,input image is-  "seed" or "target"
            mask: binary string, "no-mask" or "mask" (inherit mask from __init__)
            channel: 0,1,2 (rgb)
            :return
            3 channel stack file 8bit image with a
        '''
        input_gs_image = self.input_image
        input_gs_image = (input_gs_image / input_gs_image.max()) * 255
        ch2_u8 = np.uint8(input_gs_image)
        rgb_input_img = np.zeros((np.shape(ch2_u8)[0], np.shape(ch2_u8)[1], 3), dtype=np.uint8)
        rgb_input_img[:, :, 0] = ch2_u8
        rgb_input_img[:, :, 1] = ch2_u8
        rgb_input_img[:, :, 2] = ch2_u8
        mask_input = self.mask_roi
        if len(np.unique(mask_input)) > 1:
            bin_mask = Compsite_display.binary_frame_mask(ch=self.input_image,mask=self.mask_roi)
            bin_mask = np.where(bin_mask == 1, True, False)
            if self.channel is not None:
                rgb_input_img[bin_mask > 0, self.channel] = 255
            else:
                rgb_input_img[bin_mask > 0, 2] = 255
        return rgb_input_img

    def measure_properties(self):
        prop_names = [
            "label",
            "area",
            "eccentricity",
            "euler_number",
            "extent",
            "feret_diameter_max",
            "inertia_tensor",
            "inertia_tensor_eigvals",
            "moments",
            "moments_central",
            "moments_hu",
            "moments_normalized",
            "orientation",
            "perimeter",
            "perimeter_crofton",
            # "slice",
            "solidity"
        ]
        table_prop = measure.regionprops_table(
            self.mask_roi, intensity_image=self.input_image, properties=prop_names
        )
        return table_prop

    def display_image_label(self, table, font_select, font_size, label_draw = None, intensity = 1):
        '''
        table: table of objects measure
        label_draw: 'table index' or feature selected from table (label)
        font_select: copy font to the working directory ("DejaVuSans.ttf" eg for mac, "arial.ttf" for windows no need to copy)
        font_size: 4 is nice size
        intensity: brighter image
        return:
        PIL_image: 16 bit mask rgb of the labeled image
        '''
        PIL_image = Image.fromarray(self.rgb_out * intensity)
        if table is None:
            raise ValueError("Table is missing")
        if len(table) < 2:
            raise ValueError("No object detected")
        #table['label'] = range(2, len(table) + 2)
        # round
        info_table = table.round({'centroid-0': 0, 'centroid-1': 0})
        #info_table = info_table.reset_index(drop=True)
        draw = ImageDraw.Draw(PIL_image)
        # use a bitmap font
        font = ImageFont.truetype(font_select, font_size)
        if label_draw is None:
            # display label
            sel_lable =  info_table.index.values.tolist()
        else:
            if label_draw in [col for col in info_table.columns]:
                sel_lable =  info_table[label_draw].tolist()
            else:
                raise ValueError("Feature is not selected")
        for i,label in enumerate(sel_lable):
            draw.text((info_table.loc[i, "centroid-1"].astype('int64'), info_table.loc[i, "centroid-0"].astype('int64')),
                      str(label), 'red', font=font)
        # contrast = ImageEnhance.Contrast(PIL_image)
        # contrast_enhanced_img = contrast.enhance(intensity)
        return PIL_image

    @staticmethod
    def enhanceImage(rgb_input_img,intensity):
        '''
        :param rgb_input_img: 3 channel stack file 8bit image with a
        :param intensity: brighter image
        :return: pillow object image
        '''
        PIL_image = Image.fromarray(rgb_input_img * intensity)
        contrast = ImageEnhance.Contrast(PIL_image)
        contrast_enhanced_img = contrast.enhance(intensity)
        return contrast_enhanced_img









import tifffile as tfi
import numpy as np
from skimage.filters import threshold_local
from skimage.util import img_as_ubyte
from scipy.ndimage.morphology import binary_opening
import skimage.morphology as sm
from skimage.segmentation import watershed
from skimage import measure
import skimage
import os
import pandas as pd
from scipy.ndimage.morphology import binary_fill_holes
from PIL import Image, ImageDraw,ImageFont
import matplotlib.pyplot as plt
from AIPyS.supportFunctions.AIPS_functions import rgbTograyscale

class AIPS():
    '''
    Initiate the AIPS object
    Parameters
    ----------
        Image_name: (str)
        path: (str)
    '''
    def __init__(self, Image_name=None, path=None, inputImg = None):
        self.Image_name = Image_name
        self.path = path
        self.inputImg = inputImg
        if Image_name and path:
            input = tfi.imread(os.path.join(self.path, self.Image_name))
            imType = str(input.dtype)
            if not ('uint8' in imType):
                input = img_as_ubyte(input)
            if len(np.shape(input)) < 3:
                input = input.reshape(1, np.shape(input)[0], np.shape(input)[1])
            # stack channel is last
            elif np.shape(input)[0] > 10:
                input = rgbTograyscale(input)
            self.inputImg = input

    def imageMatrix(self):
       return self.inputImg

class Segmentation(AIPS):
    '''
    Parametric object segmentation
    '''
    def __init__(self,Image_name,path,ch_=None,rmv_object_nuc=None, block_size=None, offset=None, clean = None, seed_out = None, out_target = None):
        super().__init__(Image_name,path)
        self.ch_ = ch_
        self.rmv_object_nuc = rmv_object_nuc
        self.block_size = block_size
        self.offset = offset
        self.clean = clean
        self.seed_out = seed_out
        self.out_target = out_target
    def seedSegmentation(self):
        '''
        Parameters
        ----------
        ch_: int
            Channel selected
        block_size : float
            Detect local edges 1-99 odd
        offset : float
            Detect local edges 0.001-0.9 odd
        rmv_object_nuc : float
            percentile of cells to remove, 0.01-0.99
        clean : int
            opening operation, matrix size must be odd
        Returns
        -------
        nmask2 : float
            local threshold binary map (eg nucleus)
        nmask4 : float
            local threshold binary map post opening (eg nucleus)
        sort_mask : img
            RGB segmented image output first channel for mask (eg nucleus)
        sort_mask_bin : img
            Binary
        '''
        #
        ch = self.inputImg[self.ch_]
        nmask = threshold_local(ch, self.block_size, "mean", self.offset)
        blank = np.zeros_like(ch)
        # empty dictionary
        out = {'nmask2': blank, 'nmask4': blank, 'sort_mask': blank,
                'tabale_init': None, 'table': None}
        nmask2 = ch > nmask
        if self.clean % 2 == 0:
            out['nmask2'] = nmask2
            self.seed_out = out
            return self.seed_out
        # remove small object according to the matrix size store in clean variable
        nmask3 = binary_opening(nmask2, structure=np.ones((self.clean,self.clean))).astype(np.uint8)
        nmask4 = binary_fill_holes(nmask3)
        # Bin to ROI
        label_objects = sm.label(nmask4, background=0)
        # Calculate object size
        info_table = pd.DataFrame(
            measure.regionprops_table(
                label_objects,
                intensity_image=ch,
                properties=['area', 'label', 'coords', 'centroid'],
            )).set_index('label')
        table_init = info_table
        # remove size by quantile input
        test = info_table[info_table['area'] < info_table['area'].quantile(q=self.rmv_object_nuc)]
        sort_mask = label_objects
        if len(test) > 0:
            x = np.concatenate(np.array(test['coords']))
            sort_mask[tuple(x.T)[0], tuple(x.T)[1]] = 0
            sort_mask_bin = np.where(sort_mask > 0, 1, 0)
            out = {'nmask2': nmask2, 'nmask4': nmask4, 'sort_mask': sort_mask,
                    'sort_mask_bin': sort_mask_bin, 'tabale_init': table_init, 'table': test}
            self.seed_out = out
            return self.seed_out
        else:
            out = {'nmask2': nmask2, 'nmask4': nmask4, 'sort_mask': blank,
                    'sort_mask_bin': blank,'tabale_init': table_init, 'table': test}
            self.seed_out = out
            return self.seed_out

    def cytosolSegmentation(self, ch2_,block_size_cyto,offset_cyto,global_ther,rmv_object_cyto,rmv_object_cyto_small,remove_borders=False):
        """
        Parameters
        ----------
        ch: img
            Input image (tifffile image object)
        ch2: img
            Input image (tifffile image object)
        sort_mask: img
            RGB segmented image output first channel for mask (eg nucleus)
        sort_mask_bin img
            Binary
        block_size_cyto: int
            Detect local edges 1-99 odd
        offset_cyto: float
            Detect local edges 0.001-0.9 odd
        global_ther: float
            Percentile
        rmv_object_cyto: float
            percentile of cells to remove, 0.01-0.99
        rmv_object_cyto_small: float
            percentile of cells to remove, 0.01-0.99
        remove_border: bool
            binary, object on border of image to be removed
        Returns
        -------
        nmask2: img
            local threshold binary map (eg nucleus)
        nmask4: img
            local threshold binary map post opening (eg nucleus)
        sort_mask: img
            RGB segmented image output first channel for mask (eg nucleus)
        cell_mask_2: img
            local threshold binary map (eg cytoplasm)
        combine: img
            global threshold binary map (eg cytoplasm)
        sort_mask_syn: img
            RGB segmented image output first channel for mask (eg nucleus) sync
        mask_unfiltered: img
            Mask before filtering object size
        cseg_mask: img
            RGB segmented image output first channel for mask (eg nucleus)
        cseg_mask_bin: img
            Binary mask
        test: data_frame
            Area table seed
        info_table: data_frame
            Area table cytosol synchronize
        table_unfiltered: data_frame
            Table before remove large and small objects
        """
        # load matrix from seedSegmentation module
        sort_mask = self.seed_out['sort_mask']
        sort_mask_bin = self.seed_out['sort_mask_bin']
        # gaussian threshold
        ch2 = self.inputImg[ch2_]
        ther_cell = threshold_local(ch2, block_size_cyto, "gaussian", offset_cyto)
        blank = np.zeros_like(ch2)
        # initiate output
        outTarget = {'cell_mask_1': blank,'combine': blank,'sort_mask_sync':blank,'mask_unfiltered':blank,'cseg_mask': blank,'cseg_mask_bin':blank,'info_table': None,'table_unfiltered':None}
        cell_mask_1 = ch2 > ther_cell
        cell_mask_2 = binary_opening(cell_mask_1, structure=np.ones((self.clean,self.clean))).astype(np.float64)
        # global threshold
        quntile_num = np.quantile(ther_cell, global_ther)
        cell_mask_3 = np.where(ther_cell > quntile_num, 1, 0)
        # in-case of segmentation failed
        if np.sum(cell_mask_3)==0:
            outTarget["cell_mask_1"] = cell_mask_2
            self.out_target = outTarget
            return  self.out_target
        combine = cell_mask_2
        combine[cell_mask_3 > combine] = cell_mask_3[cell_mask_3 > combine]
        combine[sort_mask_bin > combine] = sort_mask_bin[sort_mask_bin > combine]
        # using seedSegmentation module matrix as seed and watershed for outline
        cseg = watershed(np.ones_like(sort_mask_bin), sort_mask, mask=cell_mask_2)
        csegg = watershed(np.ones_like(sort_mask), cseg, mask=combine)
        info_table = pd.DataFrame(
            measure.regionprops_table(
                csegg,
                intensity_image=ch2,
                properties=['area', 'label', 'centroid','coords'],
            )).set_index('label')
        ############# remove large object ################
        cseg_mask = csegg
        table_unfiltered = info_table
        test1 = info_table[info_table['area'] > info_table['area'].quantile(q = rmv_object_cyto)]
        if len(test1) > 0:
            x = np.concatenate(np.array(test1['coords']))
            cseg_mask[tuple(x.T)[0], tuple(x.T)[1]] = 0
        else:
            cseg_mask = cseg_mask
        ############# remove small object ################
        test2 = info_table[info_table['area'] < info_table['area'].quantile(q = rmv_object_cyto_small)]
        if len(test2) > 0:
            x = np.concatenate(np.array(test2['coords']))
            cseg_mask[tuple(x.T)[0], tuple(x.T)[1]] = 0
        else:
            cseg_mask = cseg_mask
        # sync seed mask with cytosol mask
        if remove_borders:
            y_axis = np.shape(ch2)[0]
            x_axis = np.shape(ch2)[1]
            empty_array = np.zeros(np.shape(ch2))
            empty_array[0:1, 0:y_axis] = cseg_mask[0:1, 0:y_axis]  # UP
            empty_array[y_axis - 1:y_axis, 0:y_axis] = cseg_mask[y_axis - 1:y_axis, 0:y_axis]  # down
            empty_array[0:x_axis, 0:1] = cseg_mask[0:x_axis, 0:1]
            empty_array[0:x_axis, y_axis - 1:y_axis] = cseg_mask[0:x_axis, y_axis - 1:y_axis]  # left
            u, indices = np.unique(empty_array[empty_array > 0], return_inverse=True) #u is unique values greater then zero
            remove_border_ = list(np.int16(u))
            for i in list(remove_border_):
                cseg_mask = np.where(cseg_mask == i, 0, cseg_mask)
            info_table = pd.DataFrame(
                measure.regionprops_table(
                    cseg_mask,
                    intensity_image=ch2,
                    properties=['area', 'label', 'centroid'],
                )).set_index('label')
        else :
            if len(info_table) > 1:
                info_table = pd.DataFrame(
                    measure.regionprops_table(
                        cseg_mask,
                        intensity_image=ch2,
                        properties=['area', 'label', 'centroid'],
                    )).set_index('label')
            else:
                self.out_target = outTarget
                return self.out_target
        len_unfiltered = len(table_unfiltered)
        len_test1 = len(table_unfiltered.drop(test1.index))
        len_test2 = len(table_unfiltered.drop(test2.index))
        dict_object_table = {'Start':len_unfiltered,"remove large objects":len_test1,"remove small objects":len_test2}
        table_object_summary = pd.DataFrame(dict_object_table,index=[0])
        # set mask for seed and target
        sort_mask_bin = np.where(sort_mask > 0, 1, 0)
        cseg_mask_bin = np.where(cseg_mask > 0, 1, 0)
        combine_namsk = np.where(sort_mask_bin + cseg_mask_bin > 1, sort_mask, 0)
        # dictionary update
        outTarget['combine'] = combine
        outTarget['sort_mask_sync'] = combine_namsk
        outTarget['mask_unfiltered'] = cseg
        outTarget['cseg_mask'] = cseg_mask
        outTarget['cseg_mask_bin'] = cseg_mask_bin
        outTarget['info_table'] = info_table
        outTarget['table_unfiltered'] = table_unfiltered
        self.out_target = outTarget
        return self.out_target


class AIPS_Cyto_Global:
    '''
    Initiate the AIPS object
    Parameters
    ----------
        Image_name: (str)
        path: (str)
        block_size_cyto: int
            Detect local edges 1-99 odd
        offset_cyto: float
            Detect local edges 0.001-0.9 odd
        global_ther: float
            Percentile
        clean : int
            opening operation, matrix size must be odd
    
    Returns
    -------
        combine: img
            global threshold binary map (eg cytoplasm)
    '''
    def __init__(self, Image_name=None, block_size_cyto = 11, offset_cyto = 0.1, global_ther = 0.1, clean  = 5,channel = 2,bit8 = 255,ci = 1):
        self.Image_name = Image_name #(glob)
        self.block_size_cyto = block_size_cyto
        self.offset_cyto = offset_cyto
        self.global_ther = global_ther
        self.clean = clean
        self.channel = channel
        self.bit8 = bit8
        self.ci = ci
        self.ch2 = self.loadFile()
        self.mask = self.cytosolSegmentation_()
        self.rgb_input_img = self.disply_compsite()
        
    def loadFile(self):
        ch2 = skimage.io.imread(self.Image_name)
        return ch2
    
    def cytosolSegmentation_(self):
        # load matrix from seedSegmentation module
        ch2 = self.ch2
        ther_cell = threshold_local(ch2, self.block_size_cyto, "gaussian", self.offset_cyto)
        blank = np.zeros_like(ch2)
        cell_mask_1 = ch2 > ther_cell
        cell_mask_2 = binary_opening(cell_mask_1, structure=np.ones((self.clean,self.clean))).astype(np.float64)
        # global threshold
        quntile_num = np.quantile(ther_cell, self.global_ther)
        cell_mask_3 = np.where(ther_cell > quntile_num, 1, 0)
        # in-case of segmentation failed
        if np.sum(cell_mask_3)==0:
            outTarget["cell_mask_1"] = cell_mask_2
            self.out_target = outTarget
            return  self.out_target
        combine = cell_mask_2
        combine[cell_mask_3 > combine] = cell_mask_3[cell_mask_3 > combine]
        return combine

    def disply_compsite(self):
        '''
        bit8 color assign 255 is very bright
        '''
        input_gs_image = self.ch2
        input_gs_image = input_gs_image*self.ci
        input_gs_image = (input_gs_image / input_gs_image.max()) * 255
        ch2_u8 = np.uint8(input_gs_image)
        rgb_input_img = np.zeros((np.shape(ch2_u8)[0], np.shape(ch2_u8)[1], 3), dtype=np.uint8)
        rgb_input_img[:, :, 0] = ch2_u8
        rgb_input_img[:, :, 1] = ch2_u8
        rgb_input_img[:, :, 2] = ch2_u8
        rgb_input_img[self.mask > 0, self.channel] = self.bit8
        return rgb_input_img
        


import numpy as np
from skimage import measure, morphology
from skimage.transform import resize
from skimage import io
import skimage
from skimage.util import img_as_ubyte
import pandas as pd
from random import randint
from cellpose import models, core
from skimage import img_as_ubyte
from skimage.draw import disk
from skimage import measure
from skimage import exposure
from PIL import Image, ImageDraw,ImageFont
import matplotlib.pyplot as plt
from IPython.display import HTML
from matplotlib.animation import FuncAnimation
import os
import re
import random
import string
from IPython.display import clear_output
import cv2
import glob
import pdb
import seaborn as sns
# from tifffile import imread
from AIPyS.supportFunctions import AIPS_module as am
from AIPyS.cellpose import AIPS_cellpose as AC
from AIPyS.supportFunctions.AIPS_module import AIPS_Cyto_Global
import pdb

class Granularity_cellprofiler:
    def __init__(self, image_list, diameter_cp = None, model_type="cyto",channels = [0,0]):
        '''
        image_list = glob.glob(data_dir + '*tif')
        '''
        self.image_list = image_list
        self.images = self.loadingImages()
        self.diameter_cp = diameter_cp
        self.model_type = model_type
        self.channels = channels
    
    def loadingImages(self):
        return [skimage.io.imread(image) for image in self.image_list]

    def openingOperation(self,kernel,image):
        '''
        :param: kernel (int), size of filter kernel
        :return: opening operation image
        '''
        selem = morphology.disk(kernel, dtype=bool)
        eros_pix = morphology.erosion(image, footprint=selem)
        imageOpen = morphology.dilation(eros_pix, footprint=selem)
        return imageOpen

    def cellpose_segmantation(self, image_input):
        use_GPU = core.use_gpu()
        model = models.Cellpose(gpu=use_GPU, model_type=self.model_type)
        mask, _, _ , _ = model.eval(image_input, diameter=self.diameter_cp, flow_threshold=None, channels=[0,0])
        return mask

    def openingOperation(self,image_input,kernel):
        '''
        :param: kernel (int), size of filter kernel
        :return: opening operation image
        '''
        selem = morphology.disk(kernel, dtype=bool)
        eros_pix = morphology.erosion(image_input, footprint=selem)
        imageOpen = morphology.dilation(eros_pix, footprint=selem)
        return imageOpen

    def normalize(self,image_input,mean,std):
        return (image_input - mean)/std

class ImageToVideo:
    def __init__(self,origFolder,imageNames,outFolder,alpha = 1.5, beta = 0,displayPrograss = True):
        self.origFolder = origFolder
        self.imageNames = imageNames
        self.outFolder = outFolder
        self.alpha = alpha
        self.beta = beta
        self.displayPrograss = displayPrograss
        self.loadImages()
    
    def extract_mean_sd(self,filename):
        mean = re.search('mean_(.*?)_', filename)
        if mean:
            mean = float(mean.group(1))
        sd = re.search('sd_(.*?)_', filename)
        if sd:
            sd = float(sd.group(1))
        return mean, sd

    def addText(self,image_input,text):
        return cv2.putText(image_input, f'mean: {text[0]} \n sd: {text[1]}', (10,50), cv2.FONT_HERSHEY_SIMPLEX, 1, (255,255,255), 2, cv2.LINE_AA)
    
    def rgbImge(self,image_input):
        input_gs_image = (image_input / image_input.max()) * 255
        ch2_u8 = np.uint8(input_gs_image)
        rgb_input_img = np.zeros((np.shape(ch2_u8)[0], np.shape(ch2_u8)[1], 3), dtype=np.uint8)
        rgb_input_img[:, :, 0] = ch2_u8
        rgb_input_img[:, :, 1] = ch2_u8
        rgb_input_img[:, :, 2] = ch2_u8
        pilimage = Image.fromarray(rgb_input_img)
        return rgb_input_img
    
    def imagesTovideo(self,files,imageName):
        frame = skimage.io.imread(files[0])
        frame = self.rgbImge(frame)
        width,height,layers = frame.shape
        video = cv2.VideoWriter(os.path.join(self.outFolder,imageName + '_out.avi'), cv2.VideoWriter_fourcc(*'MJPG'),10.0,(width,height),  isColor = True)  
        for file in files:
            mean, sd = self.extract_mean_sd(file)
            frame = io.imread(file)
            frame = self.rgbImge(frame)
            # add text:
            self.addText(frame,(mean,sd))
            #frame = ((frame - frame.min()) / (frame.ptp()) * 255).astype(np.uint8)
            video.write(frame)
        video.release()
    
    def loadImages(self):
        if self.displayPrograss:
           clear_output(wait=True)
        for i,imageName in enumerate(self.imageNames):
            files = glob.glob(self.origFolder + imageName + '/ImageSplit/*tif')    
            #pdb.set_trace()
            self.imagesTovideo(files,imageName)
            if self.displayPrograss:
                print('-' * i, end = '\r')


class ImageExploration(Granularity_cellprofiler,ImageToVideo):
    '''
    Explore several objects per image and intensity to determine sliding window size.
    '''
    def __init__(self,image_list,diameter_cp):
        super().__init__(image_list,diameter_cp)

    def getImageDetails(self):
        singleImage = self.images[np.random.randint(0,len(self.images))]
        return print(f'Image size: {singleImage.shape}, /n Image type: {singleImage.dtype}')
        
    def getMeanStd(self):
        '''
        Get the images mean and standard deviation
        '''
        im_inten = []
        for i in range(len(self.images)):
            im_inten.append(np.mean(self.images[i]))
        return print(f'Image mean: {np.mean(im_inten)}, /n standard deviation: {np.std(im_inten)}')

    
    def getCellPerImage(self):
        '''
        Knowing the diameter_cp number helps for cleaning the mask. for example diameter_cp = 100. 
        return the mean number of cells per image list, helps decide running window size
        '''
        cellNum = []
        for i in range(len(self.images)):
            mask = self.cellpose_segmantation(self.images[i])
            cellNum.append(len(np.unique(mask)))
        return print(f'The mean number of cells per image: {np.mean(cellNum)}')
    
    
    def kernelLossVideo_HTML(self,start_kernel = 2 , end_karnel = 20, kernel_size=10, ci = 1):
        '''
        ci - image intensity 
        return video showing image lost
        '''
        kernels = np.linspace(start_kernel, end_karnel, kernel_size, endpoint=True, dtype=int)
        singleImage = self.images[np.random.randint(0,len(self.images))]
        images = [self.openingOperation(singleImage, kernel) for kernel in kernels]
        fig, ax = plt.subplots()

        ax.axis('off')
        
        # Display the first frame
        im = ax.imshow(images[0]*ci, cmap='gray')
        
        # Function to update figure
        def update(i):
            # Update the image data
            im.set_array(images[i]*ci)
        
        # Create animation
        ani = FuncAnimation(fig, update, frames=len(images), interval=500) # 500ms between frames
        HTML(ani.to_jshtml())

    def kernelLossVideo(self,pathout = '',output_name = 'kernelLoss.avi',start_kernel = 2 , end_karnel = 20, kernel_size=10, ci = 1):
        '''
        ci - image intensity 
        return video showing image lost
        '''
        kernels = np.linspace(start_kernel, end_karnel, kernel_size, endpoint=True, dtype=int)
        files = self.image_list[np.random.randint(0,len(self.image_list))]
        singleImage = skimage.io.imread(files)
        images = [self.openingOperation(singleImage, kernel) for kernel in kernels]
        frame = self.rgbImge(images[0])
        width,height,layers = frame.shape
        video = cv2.VideoWriter(os.path.join(pathout,output_name), cv2.VideoWriter_fourcc(*'MJPG'),1.0,(width,height),  isColor = True)
        #pdb.set_trace()
        for image in images:
            mean,sd = np.mean(image), np.std(image)
            frame = self.rgbImge(image)
            # add text:
            #self.addText(frame,(mean,sd))
            video.write(frame)
        video.release()


class RunningWindow(Granularity_cellprofiler):
    '''
    Split the image according to the window size to identify the smallest image size with the highest possibility of identifying the phenotypic cell. 
    '''
    def __init__(self,image_list,windowSize,rnadomWindows):
        self.windowSize = windowSize # for example split image to 6 slides
        self.rnadomWindows = rnadomWindows # for example add 3 more random slides
        super().__init__(image_list)

    def runningWin(self,image_input):
        '''
        creates a running window sized by splitting the image to equal size images
        '''
        if not isinstance(self.windowSize, (int, float)):
                raise ValueError("window size is missing")
        if not isinstance(self.rnadomWindows, (int, float)):
            raise ValueError("rando crop generator is missing choose a number")
        minAxsis = min(image_input.shape[0],image_input.shape[1])
        winSize = self.windowSize
        while minAxsis % winSize != 0:
            minAxsis -= 1
        stepSize =  int(minAxsis/winSize)
        print(f'max image size: {minAxsis}, step size: {stepSize}')
        imageSlice = []
        for i in range(0,minAxsis,stepSize):
            for j in range(0,minAxsis,stepSize):
                imageSlice.append(image_input[i:i+stepSize,j:j+stepSize])
        # adding rando slices
        rnadomWindows = [1]*self.rnadomWindows
        range_xy = 0+stepSize,minAxsis-250
        while rnadomWindows:
            x = np.random.randint(low = range_xy[0],high = range_xy[1],size =1).tolist()[0]
            imageSlice.append(image_input[x:x+stepSize,x:x+stepSize])    
            rnadomWindows.pop()
        return imageSlice

class Granularity(RunningWindow):
    def __init__(self,pathOrig,meanInten,stdInten,kernel,image_list,windowSize,rnadomWindows,displayPrograss = True):
        self.pathOrig = pathOrig
        self.meanInten = meanInten
        self.stdInten = stdInten
        self.kernel = kernel
        self.displayPrograss = displayPrograss
        super().__init__(image_list,windowSize,rnadomWindows)

    def generate_random_string(self,length):
        letters = string.ascii_letters  # this includes both lowercase and uppercase letters
        result_str = ''.join(random.choice(letters) for i in range(length))
        return result_str
    
    def getMeanSD(self,input_image):
        return np.round(np.mean(input_image),3),np.round(np.std(input_image),3)
    
    def imageMeasure(self,image_input):
        meanInten = self.meanInten
        stdInten = self.stdInten
        image_norm = self.normalize(image_input,meanInten,stdInten)
        image_gran_cal = self.openingOperation(image_input = image_norm, kernel = self.kernel)
        mean,sd = self.getMeanSD(image_gran_cal)
        return mean,sd
    
    def granCalc(self):
        '''
        Save normal mean and sd intensity of normal images and crop image
        '''
        if not isinstance(self.meanInten, (int, float)):
                raise ValueError("calculate mean image intensity")
        if not isinstance(self.stdInten, (int, float)):
            raise ValueError("calculate sd image intensity")
        for image in self.images:
            # creat folder
            image_name = self.generate_random_string(7)
            tempFolder = os.path.join(self.pathOrig,image_name)
            ImageOrig = os.path.join(tempFolder,'ImageOrig')
            ImageSplit = os.path.join(tempFolder,'ImageSplit')
            os.makedirs(tempFolder,exist_ok=True)
            os.makedirs(ImageOrig,exist_ok=True)
            os.makedirs(ImageSplit,exist_ok=True)
            mean,sd = self.imageMeasure(image)
            skimage.io.imsave(os.path.join(ImageOrig,image_name + f'_mean_{mean}' + f'_sd_{sd}_' +'.tif'), image)
            imageSlice = self.runningWin(image_input = image)
            if self.displayPrograss:
                clear_output(wait=True)
                print(f'Image Name: {image_name}') 
            for i,imageCrop in enumerate(imageSlice):
                mean,sd = self.imageMeasure(imageCrop)
                skimage.io.imsave(os.path.join(ImageSplit,image_name + f'_mean_{mean}' + f'_sd_{sd}_' +'.tif'), imageCrop)
                if self.displayPrograss:
                    print('-' * i, end = '\r')

class GranulaityMesure(AIPS_Cyto_Global):
    def __init__(self,Image_name, block_size_cyto, offset_cyto, global_ther, clean,start_kernel = 2,end_karnel = 50,kernel_size = 20,outputTableName = None,outPath = None):
        self.start_kernel = start_kernel
        self.end_karnel = end_karnel
        self.kernel_size = kernel_size
        self.outputTableName = outputTableName
        self.outPath = outPath
        super().__init__(Image_name, block_size_cyto, offset_cyto, global_ther, clean)
        self.granularityMesure_globalSegmant()
    
    def openingOperation(self,kernel,image):
        '''
        Parameters
        ---------- 
            kernel: int, 
                size of filter kernel
        return
        ------
            opening operation image
        '''
        selem = morphology.disk(kernel, dtype=bool)
        eros_pix = morphology.erosion(image, footprint=selem)
        imageOpen = morphology.dilation(eros_pix, footprint=selem)
        return imageOpen
    
    
    def loopLabelimage(self,deploy = False ):
        '''
        return: dataframe of image intensity per segmented cell per image
        '''
        # returns column wise table
        open_vec = np.linspace(self.start_kernel, self.end_karnel, self.kernel_size, endpoint=True, dtype=int)
        if deploy:
            limit = 1
            open_vec = [open_vec[0],open_vec[-1]]
        else:
            limit = 3
        if len(open_vec) < limit:
            mesg = "increase kernel size"
            raise ValueError(mesg)
        mean_int = [] # image mean signal intensity
        sd_int = [] # image mean signal intensity
        for opning in open_vec:
            openImage = self.openingOperation(kernel = opning,image = self.ch2)
            mean_int.append(np.mean(openImage[self.mask>0]))
            sd_int.append(np.std(openImage[self.mask>0]))
        mean_int_norm = [int/mean_int[0] for int in mean_int]
        table = pd.DataFrame({'kernel':open_vec,'mean_decay':mean_int_norm,'mean':mean_int,'sd':sd_int })
        return table
    
    def granularityMesure_globalSegmant(self,saveMode = True):
        '''
        function description:
            - [] cell segmented using global threshold 
            - [] granularity measure
        output:
            Segmented image composition with area label
            - [] Area histogram plot
            - [] Segmented image composition after removal of small objects
            - [] plot of Mean intensity over opening operation (granularity spectrum)
        Parameters
        ----------
        clean: int
            remove object bellow the selected area size
        classLabel: int
            assign label
        file: str
        path: str
        outPath: str
        outputTableName: str
            e.g. "outputTableNorm.csv"
        saveMode: bool
            save returns
        Notes
        -----
        required single channel tif
        '''
        img = self.ch2 # from AIPS_Cyto_Global
        if len(img.shape) > 2:
            raise ValueError("required single channel tif")
        mask = self.mask 
        compImage = self.rgb_input_img #pil image
        granData = self.loopLabelimage()
        if saveMode:
            skimage.io.imsave(os.path.join(self.outPath,self.outputTableName + ".png"),compImage)
            granData.to_csv(os.path.join(self.outPath, self.outputTableName + ".png"))
        fig, ax = plt.subplots()
        sns.lineplot(data=granData, x="kernel", y="mean_decay").set(title='Granularity spectrum plot')
        

            

    



        


#%%
import dash
from dash import html
from dash import dash_table, dcc
from dash.dependencies import Input, Output
import dash_daq as daq
#import dash_core_components as dcc
from dash import State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import pandas as pd
import sys
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import seaborn as sns
import os
import re
import glob
import random
from sklearn import preprocessing
from tqdm import tqdm
import plotly.express as px
import shutil

# dataPath = r'D:\Gil\images\pex_project\02142024\inc_origImage\imageSequence\data'
# imagePath = r'D:\Gil\images\pex_project\02142024\inc_origImage\imageSequence\images'
def image_labeling(imagePath,dataPath):
  # normlise intensity and sd and mask area by origImage.

  df_Orig = pd.read_csv(os.path.join(dataPath,'imageseq_data.csv'))
  # copy data before labeling to new folder 
  df_Nolabel = os.path.join(dataPath,"prelabel_table")
  if not os.path.exists(df_Nolabel):
    os.makedirs(df_Nolabel, exist_ok=True)
    shutil.copy('imageseq_data.csv', df_Nolabel)
  
  def reMeanSD(data,imageName):
      mean = data.loc[data['name'].str.contains(imageName),'intensity'].values
      sd = data.loc[data['name'].str.contains(imageName),'sd'].values
      return mean,sd
      
  def normIntenSd(df,dfOrig,sel_name):
      curr_name = re.sub('_.*','',sel_name)
      mean,sd = reMeanSD(dfOrig,curr_name)
      mean_list_curr = df.loc[df['name'].str.contains(curr_name),'intensity'].values
      sd_list_curr = df.loc[df['name'].str.contains(curr_name),'sd'].values
      df.loc[df['name'].str.contains(curr_name),'signal_norm'] = mean_list_curr/mean
      df.loc[df['name'].str.contains(curr_name),'sd_norm'] = sd_list_curr/sd
      
      
  df = pd.read_csv(os.path.join(dataPath,'imageseq_data.csv'))
  float_cols = df.select_dtypes(include=['float']).columns
  df[float_cols] = df[float_cols].apply(lambda x: np.round(x, 4))
  image_names = [re.sub("_.*","",name) for name in set(df.name.tolist())]
  
  df['signal_norm'] = 0
  df['sd_norm'] = 0
  # normalize data frame- traverse by unique image name and add normalize intensity and sd to df 
  # for imag_name_curr in image_names:
  #     normIntenSd(df,df_Orig,imag_name_curr)

  np.random.seed(42423)
  df = df.sample(frac=1).reset_index(drop=True)
  #df['name'] = df['name'].apply(lambda x: f"{x}.png")
  first_image_name = df.iloc[0]['name']
  first_image_path = f"{imagePath}/{first_image_name}"
  savedContainer = {'name':[], 'ratio':[], 'intensity':[], 'sd':[], 'maskArea':[], 'label':[]}
  image_names = [re.sub("_.*","",name) for name in set(df.name.tolist())]

  def openImageToPX(imageName):
    im_pil = Image.open(imageName)
    img = px.imshow(im_pil,binary_string=True, binary_backend="png", width=650, height=650,binary_compression_level=9).update_xaxes(showticklabels=False).update_yaxes(showticklabels = False)
    return img

  app = dash.Dash(__name__,prevent_initial_callbacks=True,suppress_callback_exceptions=True)
  application = app.server
  app.layout = html.Div([
        html.Div([
              dash_table.DataTable(id='table',
              columns=[{"name": i, "id": i,}
                      for i in df.columns if i in ['name', 'ratio', 'label']],
              data=df.to_dict('records'),
              row_selectable="single",
              editable = True,
              page_size=10,
              page_current = 0,
              page_action='native',
              row_deletable=True,
              cell_selectable=True),
               html.Div([
                daq.BooleanSwitch(id='my-boolean-switch',
                                label="Null label remove",
                                labelPosition="top",
                                on=False),
                dbc.Button('Save data', id='save-button', color="danger", n_clicks=0, active=True),
                dcc.Loading(html.Div(id='load-csv-file'), type="circle", style={'height': '100%', 'width': '100%'}),
                          ],style={'display': 'flex', 'flex-direction': 'row'})
          ],style={'flex': '1', 'display': 'flex', 'flex-direction': 'column'}),
          html.Div([
            html.Div(id='img-container')
          ], style={'flex': '1'}),
          html.Div(id='temp-contain'),
  ],style={'display': 'flex', 'flex-direction': 'row'})

  @app.callback(
    Output('img-container', 'children'),
    [Input('table', 'active_cell'),
    State('table', 'page_current'),
    State('table', 'page_size')],
    prevent_initial_call=True,)
  def display_image(active_cell,page_current,page_size):
    '''
    active_cell - dictionery {'row': 0, 'column': 0, 'column_id': 'name'}
    rows - list of dict per row:  [{'Unnamed: 0': 294, 'name': 'QOnvDbK_24.png', 'ratio': 0.032,
    'intensity': 783.206, 'sd': 611.006, 'maskArea': 0.995, 'label': 0},....]
    '''
    if (active_cell is None) and (page_current == 0):
      img_name = df.loc[0,'name']
      img = openImageToPX(os.path.join(imagePath,img_name))
      return [dcc.Graph(id="disp3",figure=img)]
    elif (active_cell is None) and (page_current > 0):
      return ["select image name"]
    elif (active_cell['column_id'] == 'name') and (page_current == 0):
      img_name = df.loc[active_cell['row'],active_cell['column_id']]
      img = openImageToPX(os.path.join(imagePath,img_name))
      return [dcc.Graph(id="disp2",figure=img)]
    elif  (active_cell['column_id'] == 'name') and (page_current > 0):
      df_new = df.iloc[page_current*page_size:(page_current+1)*page_size].reset_index()
      img_name = df_new.loc[active_cell['row'],active_cell['column_id']]  # No need to append ".png"
      img = openImageToPX(os.path.join(imagePath,img_name))
      return [dcc.Graph(id="disp1",figure=img)]
    raise PreventUpdate

  @app.callback(
    Output('load-csv-file', 'children'),
    [Input('save-button', 'n_clicks'),
     Input('my-boolean-switch','on'),
    State('table', 'data')],
    prevent_initial_call=True,
  )
  def save_data(n,on,curr_table):
    # Handler for data update; includes save logic as needed
    if n is None:
      return dash.no_update
    else:
      df_curr = pd.DataFrame(curr_table).reset_index()
      if on:
        # remove all the rows with no label
        df_curr = df_curr[df_curr['label'] != 'na']
      df_curr.to_csv(os.path.join(dataPath,'imageseq_data.csv'),index=False)
      return "data is saved"
  return app




# %%



import tensorflow as tf
from tqdm import tqdm
import numpy as np
from sklearn.preprocessing import LabelEncoder
import pickle

np.random.seed(42)
import matplotlib.pyplot as plt
from tensorflow.keras.preprocessing.image import ImageDataGenerator, load_img, img_to_array, array_to_img
from tensorflow.keras.layers import Conv2D, MaxPooling2D, Flatten, Dense, Dropout
from tensorflow.keras.models import Sequential
from tensorflow.keras import optimizers

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

from keras.applications import vgg16
from keras.models import Model
import keras
import pandas as pd


from keras.layers import Conv2D, MaxPooling2D, Flatten, Dense, Dropout, InputLayer
from keras.models import Sequential
from tensorflow.keras import optimizers

import glob
import numpy as np

from keras.preprocessing.image import load_img, img_to_array, array_to_img
import os

class model_builder():
    """
    Parameters
    ----------
    IMG_DIM: tuple
        The dimensions of images to be used for training and validation.
    path_training: str
        The path to the directory containing training images.
    path_validation: str
        The path to the directory containing validation images.
    label: tuple
        binary labeling for example 'norm' and 'pheno'
    train_imgs_scaled: tf.array
        The scaled training images.
    validation_imgs_scaled:  tf.array
        The scaled validation images.
    path_model: str
        The path to the directory containing the model to be used for training.
    batch: int
        The batch size to be used for training.
    epoch: int
        The number of epochs to be used for training.
    input_shape: int
        The shape of the inputs to be used for training.
    train_labels_enc: tf.array
        The encoded labels associated with the training images.
    validation_labels_enc: tf.array
        The encoded labels associated with the validation images.
    train_imgs: image
        The training images.
    validation_imgs: image
        The validation images.
    steps_per_epoch_sel: int
        The number of steps to take per epoch during training.
    validation_steps: int
        The number of steps to take during validation.
    file_extention: str
        The file extension for the images to be used for training and validation.
    extract_size_train: int
        The size of the training set to be used for training.
    extract_size_val: int
        The size of the validation set to be used for validation.
    imbalance_train: int
        The imbalance ratio of the training set.
    imbalance_val: int
        The imbalance ratio of the validation set.
    model_name: str
        The name of the model to be used for training.
    path_checkpoints: str
        The path to the directory containing the checkpoints of the model.
    """
    def __init__(self, IMG_DIM=None, path_training=None, path_validation=None,label = None,
                train_imgs_scaled=None, validation_imgs_scaled=None, path_model=None, batch=None, epoch=None,
                input_shape=None, train_labels_enc=None, validation_labels_enc = None, train_imgs = None,
                 validation_imgs = None, steps_per_epoch_sel = None,validation_steps = None, file_extention = None,
                 extract_size_train = None,extract_size_val = None,imbalance_train = None, 
                 imbalance_val = None, model_name = None, path_checkpoints=None):
        self.IMG_DIM = IMG_DIM
        self.path_training = path_training
        self.path_validation = path_validation
        self.label = label
        self.train_imgs_scaled = train_imgs_scaled
        self.validation_imgs_scaled = validation_imgs_scaled
        self.path_model = path_model
        self.batch = batch
        self.epoch = epoch
        self.input_shape = input_shape
        self.train_labels_enc = train_labels_enc
        self.validation_labels_enc = validation_labels_enc
        self.train_imgs = train_imgs
        self.validation_imgs = validation_imgs
        self.steps_per_epoch_sel = steps_per_epoch_sel
        self.validation_steps = validation_steps
        self.file_extention = file_extention
        self.extract_size_train = extract_size_train
        self.extract_size_val = extract_size_val
        self.imbalance_train = imbalance_train
        self.imbalance_val = imbalance_val
        self.model_name = model_name
        self.path_checkpoints = path_checkpoints
        
    def display_data_distribution(self):
        train_files = glob.glob(os.path.join(self.path_training,'*.' + self.file_extention))
        validation_files = glob.glob(os.path.join(self.path_validation,'*.' + self.file_extention))
        train_labels = [self.label[0] if self.label[0] in os.path.basename(fn) else self.label[1] for fn in train_files]
        validation_labels = [self.label[0] if self.label[0] in os.path.basename(fn) else self.label[1] for fn in validation_files]
        print('number of training samples norm: {}'.format(train_labels.count(self.label[0])))
        print('number of training samples pheno: {}'.format(train_labels.count(self.label[1])))
                                                      
    def build_image__sets(self):
        # Training file generator
        train_files = glob.glob(os.path.join(self.path_training,'*.' + self.file_extention))
        if self.extract_size_train is not None:
            train_files = train_files[:self.extract_size_train]
        train_imgs = []
        # method for dealing with imbalance data, cut the norm in to minority
        if self.imbalance_train is not None:
            train_files_new = []
            count_norm = 0  
            for i in range(len(train_files)):
                if train_files[i].split('/')[7].split('_')[0].strip()=='norm':
                        if count_norm < self.imbalance_train:
                            train_files_new.append(train_files[i])
                            count_norm += 1
                else:
                    train_files_new.append(train_files[i])
            train_files = train_files_new # update variable
        print('run training')
        train_imgs = [img_to_array(load_img(train_files[i], target_size=self.IMG_DIM)) for i in tqdm(range(len(train_files)))]
        train_imgs = np.array(train_imgs)
        self.train_imgs = train_imgs
        train_labels = [self.label[0] if self.label[0] in os.path.basename(fn) else self.label[1] for fn in train_files]
        # Validation file generator
        validation_files = glob.glob(os.path.join(self.path_validation,'*.' + self.file_extention))
        if self.extract_size_val is not None:
            validation_files = validation_files[:self.extract_size_val]
        # method for dealing with imbalance data, cut the norm in to minority
        if self.imbalance_val is not None:
            validation_files_new = []
            count_norm = 0  
            for i in range(len(validation_files)):
                if validation_files[i].split('/')[7].split('_')[0].strip()=='norm':
                        if count_norm < self.imbalance_val:
                            validation_files_new.append(validation_files[i])
                            count_norm += 1
                else:
                    validation_files_new.append(validation_files[i])
            validation_files = validation_files_new # update variable
        print('run valadtion')
        validation_imgs = [img_to_array(load_img(validation_files[i], target_size=self.IMG_DIM)) for i in tqdm(range(len(validation_files)))]
        validation_imgs = np.array(validation_imgs)
        self.validation_imgs = validation_imgs
        validation_labels = [self.label[0] if self.label[0] in os.path.basename(fn) else self.label[1] for fn in validation_files]
        train_imgs_scaled = train_imgs.astype('float32')
        validation_imgs_scaled = validation_imgs.astype('float32')
        self.train_imgs_scaled = train_imgs_scaled/255
        self.validation_imgs_scaled = validation_imgs_scaled/255
        le = LabelEncoder()
        le.fit(train_labels)
        self.train_labels_enc = le.transform(train_labels)
        self.validation_labels_enc = le.transform(validation_labels)
        report = 'tarin labels:{}, train_labels_enc:{}.'.format(train_labels[10:15],self.train_labels_enc[10:15])
        return self.train_imgs_scaled, self.validation_imgs_scaled,self.train_labels_enc,\
               self.validation_labels_enc,self.train_imgs,self.validation_imgs,report

    def model_cnn_basic(self):
        model = Sequential()

        model.add(Conv2D(16, kernel_size=(3, 3), activation='relu',
                         input_shape=self.input_shape))
        model.add(MaxPooling2D(pool_size=(2, 2)))

        model.add(Conv2D(64, kernel_size=(3, 3), activation='relu'))
        model.add(MaxPooling2D(pool_size=(2, 2)))

        model.add(Conv2D(128, kernel_size=(3, 3), activation='relu'))
        model.add(MaxPooling2D(pool_size=(2, 2)))

        model.add(Flatten())
        model.add(Dense(512, activation='relu'))
        model.add(Dense(1, activation='sigmoid'))

        model.compile(loss='binary_crossentropy',
                      optimizer=optimizers.RMSprop(),
                      metrics=['accuracy'])

        model.summary()

        history = model.fit(x=self.train_imgs_scaled, y=self.train_labels_enc,
                            validation_data=(self.validation_imgs_scaled, self.validation_labels_enc),
                            batch_size=self.batch,
                            epochs=self.epoch,
                            verbose=1)

        # save the model
        os.chdir(self.path_model)
        model.save('cnn_basic.h5')
        return history

    def model_cnn_basic_Augmentation(self):
        train_datagen = ImageDataGenerator(rescale=1. / 255, zoom_range=0.3, rotation_range=50,
                                           width_shift_range=0.2, height_shift_range=0.2, shear_range=0.2,
                                           horizontal_flip=True, fill_mode='wrap', brightness_range=[0.1, 0.9])

        val_datagen = ImageDataGenerator(rescale=1. / 255)

        # applay new model with droput and aougmentation
        train_generator = train_datagen.flow(self.train_imgs, self.train_labels_enc, batch_size=self.batch)
        val_generator = val_datagen.flow(self.validation_imgs, self.validation_labels_enc, batch_size=self.batch)

        model = Sequential()

        model.add(Conv2D(16, kernel_size=(3, 3), activation='relu',
                         input_shape=self.input_shape))
        model.add(MaxPooling2D(pool_size=(2, 2)))

        model.add(Conv2D(64, kernel_size=(3, 3), activation='relu'))
        model.add(MaxPooling2D(pool_size=(2, 2)))

        model.add(Conv2D(128, kernel_size=(3, 3), activation='relu'))
        model.add(MaxPooling2D(pool_size=(2, 2)))

        model.add(Conv2D(128, kernel_size=(3, 3), activation='relu'))
        model.add(MaxPooling2D(pool_size=(2, 2)))

        model.add(Flatten())
        model.add(Dense(512, activation='relu'))
        model.add(Dropout(0.3))
        model.add(Dense(512, activation='relu'))
        model.add(Dropout(0.3))
        model.add(Dense(1, activation='sigmoid'))

        model.compile(loss='binary_crossentropy',
                      optimizer=optimizers.RMSprop(lr=1e-4),
                      metrics=['accuracy'])

        history = model.fit_generator(train_generator, steps_per_epoch=self.steps_per_epoch_sel, epochs=self.epoch,
                                      validation_data=val_generator, validation_steps=self.validation_steps,
                                      verbose=1)
        # save the model
        os.chdir(self.path_model)
        model.save('cnn_basic_Augmentation.h5')

        return history

    def model_cnn_transfer_learning_Augmentation_freez_all(self):
        vgg = vgg16.VGG16(include_top=False, weights='imagenet',
                          input_shape=self.input_shape)

        output = vgg.layers[-1].output
        output = keras.layers.Flatten()(output)
        vgg_model = Model(vgg.input, output)

        vgg_model.trainable = False
        for layer in vgg_model.layers:
            layer.trainable = False

        # show which layer are frozen
        pd.set_option('max_colwidth', -1)
        layers = [(layer, layer.name, layer.trainable) for layer in vgg_model.layers]
        pd.DataFrame(layers, columns=['Layer Type', 'Layer Name', 'Layer Trainable'])

        # flat all the fetures from training and validation set
        # for feeding the last blovk of vg16
        def get_bottleneck_features(model, input_imgs):
            features = model.predict(input_imgs, verbose=0)
            return features

        train_features_vgg = get_bottleneck_features(vgg_model, self.train_imgs_scaled)
        validation_features_vgg = get_bottleneck_features(vgg_model, self.validation_imgs_scaled)

        report = 'Train Bottleneck Features:{},  Validation Bottleneck Features:{}'.format(train_features_vgg.shape,validation_features_vgg.shape)
        '''
        now we will train with aougmentation
        Pre-trained CNN model as a Feature Extractor with Image Augmentation
        '''
        train_datagen = ImageDataGenerator(rescale=1. / 255, zoom_range=0.3, rotation_range=50,
                                           width_shift_range=0.2, height_shift_range=0.2, shear_range=0.2,
                                           horizontal_flip=True, fill_mode='wrap', brightness_range=[0.1, 0.9])

        val_datagen = ImageDataGenerator(rescale=1. / 255)

        # applay new model with droput and aougmentation
        train_generator = train_datagen.flow(self.train_imgs, self.train_labels_enc, batch_size=self.batch)
        val_generator = val_datagen.flow(self.validation_imgs, self.validation_labels_enc, batch_size=self.batch)
        

        model = Sequential()
        model.add(vgg_model)
        model.add(Dense(512, activation='relu', input_dim=self.input_shape))
        model.add(Dropout(0.3))
        model.add(Dense(512, activation='relu'))
        model.add(Dropout(0.3))
        model.add(Dense(1, activation='sigmoid'))

        model.compile(loss='binary_crossentropy',
                      optimizer=optimizers.RMSprop(learning_rate=2e-5),
                      metrics=['accuracy'])

        history = model.fit_generator(train_generator, steps_per_epoch=self.steps_per_epoch_sel, epochs=self.epoch,
                                      validation_data=val_generator, validation_steps=self.validation_steps,
                                      verbose=1)

        # save the model
        os.chdir(self.path_model)
        model.save('transfer_learning_aug_dropout_freez_all.h5')

        return history

    def model_cnn_transfer_learning_Augmentation_drop_layer_4and5(self,**kwargs):
        vgg = vgg16.VGG16(include_top=False, weights='imagenet',
                          input_shape=self.input_shape)

        output = vgg.layers[-1].output
        output = keras.layers.Flatten()(output)
        vgg_model = Model(vgg.input, output)

        vgg_model.trainable = True

        set_trainable = False
        for layer in vgg_model.layers:
            if layer.name in ['block5_conv1', 'block4_conv1']:
                set_trainable = True
            if set_trainable:
                layer.trainable = True
            else:
                layer.trainable = False

        layers = [(layer, layer.name, layer.trainable) for layer in vgg_model.layers]
        pd.DataFrame(layers, columns=['Layer Type', 'Layer Name', 'Layer Trainable'])

        train_datagen = ImageDataGenerator(rescale=1. / 255, zoom_range=0.3, rotation_range=50,
                                           width_shift_range=0.2, height_shift_range=0.2, shear_range=0.2,
                                           horizontal_flip=True, fill_mode='wrap', brightness_range=[0.1, 0.9])

        val_datagen = ImageDataGenerator(rescale=1. / 255)

        # applay new model with droput and aougmentation
        train_generator = train_datagen.flow(self.train_imgs, self.train_labels_enc, batch_size=self.batch)
        val_generator = val_datagen.flow(self.validation_imgs, self.validation_labels_enc, batch_size=self.batch)
        input_shape = self.input_shape

        model = Sequential()
        model.add(vgg_model)
        model.add(Dense(512, activation='relu', input_dim=input_shape))
        model.add(Dropout(0.3))
        model.add(Dense(512, activation='relu'))
        model.add(Dropout(0.3))
        model.add(Dense(1, activation='sigmoid'))

        learning_rate = kwargs.get('learning_rate', 1e-5)
        if 'learning_rate' in kwargs:
            learning_rate = kwargs['learning_rate']
        model.compile(loss='binary_crossentropy',
                      optimizer=optimizers.RMSprop(learning_rate=learning_rate),
                      metrics=['accuracy'])
        
        checkpoint_path = self.path_checkpoints + "cp-{epoch:04d}.ckpt"
        checkpoint_dir = os.path.dirname(checkpoint_path)
        
        cp_callback = tf.keras.callbacks.ModelCheckpoint(filepath=checkpoint_path,
                                                 save_weights_only=True,
                                                 verbose=1)
        
        history = model.fit_generator(train_generator, steps_per_epoch=self.steps_per_epoch_sel, epochs=self.epoch,
                                      validation_data=val_generator, validation_steps=self.validation_steps,
                                      verbose=1,callbacks=[cp_callback])

        os.chdir(self.path_model)
        if self.model_name:
            model.save(self.model_name)
            with open(self.model_name + '_history'  , 'wb') as file_pi:
                pickle.dump(history.history, file_pi)
        else:
            model.save('cnn_transfer_learning_Augmentation_drop_layer_4and5.h5')
            with open('history_', 'wb') as file_pi:
                pickle.dump(history.history, file_pi)
        return history






import numpy as np
import os
import skimage.io
import matplotlib.pyplot as plt
from cellpose import models, core
import pandas as pd
from scipy.stats import skew

from PIL import Image, ImageDraw,ImageFont
from skimage import img_as_ubyte
from skimage.draw import disk
from skimage import measure

import seaborn as sns
import time
from AIPyS.DataLoad import DataLoad

class AIPS_cellpose(DataLoad):
    """
    Cellpose algorithm

    Parameters
    ----------
    model_type: str
        'cyto' or model_type='nuclei'
    clean: int
        remove object bellow the selected area size
    channels: list
        channels = [0,0] # IF YOU HAVE GRAYSCALE
        channels = [2,3] # IF YOU HAVE G=cytoplasm and B=nucleus
        channels = [2,1] # IF YOU HAVE G=cytoplasm and R=nucleus
        or if you have different types of channels in each image
        channels = [[2,3], [0,0], [0,0]]
        channels = [1,1]
    diameter: int
        e.g. 50 for 20X objective
    """
    def __init__(self,model_type,channels,diameter,*args, **kwargs):
        self.model_type = model_type
        self.channels = channels
        self.diameter = diameter
        super().__init__(*args, **kwargs)
        
    def cellpose_segmentation(self, image_input):
        '''
        returns a binery mask (np array 16bit) and a table(coloumns: area	label	centroid-0	centroid-1)
        '''
        if image_input is None:
            image_input = self.img # should be single greyscale image
        use_GPU = core.use_gpu()
        model = models.Cellpose(gpu=use_GPU, model_type=self.model_type)
        start_time = time.time()
        mask, _, _, _ = model.eval(image_input, diameter=self.diameter, flow_threshold=None, channels=self.channels)
        table = pd.DataFrame(
            measure.regionprops_table(
                mask,
                intensity_image=image_input,
                properties=['area', 'label', 'centroid','intensity_mean']))
        end_time = time.time()  # Record the end time
        duration = end_time - start_time  # Calculate the duration
        if duration > 60:  # 60 seconds threshold
            raise "cellpose segmentation took too long. readjust pararmeters"
        return mask, table


import os
from pathlib import Path
import argparse
import subprocess
import sys
import h5py
import numpy as np

def unpack_dataset(dataset):
    """Handle decoding of datasets from HDF5, with special handling for arrays of strings."""
    if isinstance(dataset, np.ndarray) and dataset.dtype.kind == 'O':  # Object arrays, potentially strings
        return dataset[0].decode('utf-8')  # Decode each string from bytes
    elif isinstance(dataset, np.ndarray) and dataset.dtype.type is np.bytes_:  # Direct array of bytes
        return np.char.decode(dataset, 'utf-8')  # Decoding bytes array to strings
    elif isinstance(dataset, bytes):  # Single item as bytes
        return dataset.decode('utf-8')  # Decode single byte string to normal string
    else:
        return dataset  # Return as is for other datatypes


def check_and_prompt_parameters(option, h5filepath):
    required_params = {
        "measDia": ["Image_name"],
        "cp_seg_video": ["diameter","data_dir","imagesN", "Image_name", "videoName", "model_type", "channels", "outPath"],
        "cp_gran_video": ["outputImageSize","start_kernel", "end_karnel", "kernel_size", "extract_pixel", "resize_pixel", "videoName", "model_type", "channels", "Image_name", "outPath","diameter"],
        "cp_gran_table_gen": ["kernelGran", "w", "h", "extract_pixel", "resize_pixel", "videoName", "model_type", "channels", "Image_name", "outPath","imagesN","data_dir","trainingDataPath","diameter"],
        "dataLabeling": ["dataPath", "imagePath"],
        "data_viz": ["dataPath", "imagePath"],
        "modelBuild": ["diameter","dataPath", "imW", "imH", "thold", "extract_pixel", "resize_pixel","model_type", "channels", "Image_name", "outPath","areaSel","fractionData"],
        "deployBuild": ["diameter","kernelGran", "dataPath", "imW", "imH", "thold", "extract_pixel", "resize_pixel","model_type", "channels", "Image_name", "outPath","intercept","slope","areaSel","fractionData"]
    }
    
    
    # Opening the HDF5 file
    with h5py.File(h5filepath, 'r') as h5file:
        sub_data_dict = {}
        non_counts = 0
        missingParam = [] 
        for subkey in required_params[option]:
            try:
                value = unpack_dataset(h5file[subkey][()])
                sub_data_dict[subkey] = value
            except KeyError:
                raise f" {subkey} is required. use set_parameters --{subkey} to update parameters"
            # test if the parameter is missing
            if value=='None':
                non_counts =+ 1
                missingParam.append(subkey)
                # raise f" {subkey} is required. use set_parameters --{subkey} to update parameters"
    return sub_data_dict,non_counts,missingParam


from skimage import measure, morphology
import os
import numpy as np
import cv2

def openingOperation(kernel,image):
    '''
    Parameters
    ---------- 
        kernel: int, 
            size of filter kernel
    return
    ------
        opening operation image
    '''
    selem = morphology.disk(kernel, dtype=bool)
    eros_pix = morphology.erosion(image, footprint=selem)
    imageOpen = morphology.dilation(eros_pix, footprint=selem)
    return imageOpen


def resize(image,width,hight):
    resized = cv2.resize(image, (width,hight), interpolation = cv2.INTER_AREA)
    return resized


def imageToRGB(image,mask,ci =1):
    '''
    bit8 color assign 255 is very bright
    '''
    input_gs_image = image
    input_gs_image = input_gs_image*ci
    input_gs_image = (input_gs_image / input_gs_image.max()) * 255
    ch2_u8 = np.uint8(input_gs_image)
    rgb_input_img = np.zeros((np.shape(ch2_u8)[0], np.shape(ch2_u8)[1], 3), dtype=np.uint8)
    rgb_input_img[:, :, 0] = ch2_u8
    rgb_input_img[:, :, 1] = mask
    rgb_input_img[:, :, 2] = ch2_u8
    return rgb_input_img

def classify(a,b,Int, td):
    mu = a + b * Int
    prob = 1 / (1 + np.exp(-mu))
    return prob, prob > thold










#%%
import dash
from dash import html
from dash import dash_table, dcc
from dash.dependencies import Input, Output
#import dash_core_components as dcc
from dash import State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import pandas as pd
import sys
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import seaborn as sns
import os
import re
import glob
import random
from sklearn import preprocessing
from tqdm import tqdm
import plotly.express as px

dataPath = r'D:\Gil\images\pex_project\02142024\inc_origImage\imageSequence\data'
IMAGE_FOLDER = r'D:\Gil\images\pex_project\02142024\inc_origImage\imageSequence\images'
# normlise intensity and sd and mask area by origImage.
#%%
df_Orig = pd.read_csv(os.path.join(dataPath,'imageOrig_data.csv'))
#%%
def reMeanSD(data,imageName):
    mean = data.loc[data['name'].str.contains(imageName),'intensity'].values
    sd = data.loc[data['name'].str.contains(imageName),'sd'].values
    return mean,sd
    
def normIntenSd(df,dfOrig):
    curr_name = re.sub('_.*','',df.loc[0,'name'])
    mean,sd = reMeanSD(dfOrig,curr_name)
    mean_list_curr = df.loc[df['name'].str.contains(curr_name),'intensity'].values
    sd_list_curr = df.loc[df['name'].str.contains(curr_name),'sd'].values
    df.loc[df['name'].str.contains(curr_name),'intensity'] = mean_list_curr/mean
    df.loc[df['name'].str.contains(curr_name),'intensity'] = sd_list_curr/sd
    
    
df = pd.read_csv(os.path.join(dataPath,'imageseq_data.csv'))
float_cols = df.select_dtypes(include=['float']).columns
df[float_cols] = df[float_cols].apply(lambda x: np.round(x, 3))
normIntenSd(df,df_Orig)
np.random.seed(42423)
df = df.sample(frac=1).reset_index(drop=True)
# %%



import tifffile as tfi
import skimage.measure as sme
import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image, ImageEnhance, ImageDraw,ImageFont
# from PIL import fromarray
from numpy import asarray
from skimage import data, io
from skimage.filters import threshold_otsu, threshold_local
from skimage.morphology import convex_hull_image
import matplotlib.pyplot as plt
from scipy import ndimage as ndi
from scipy.ndimage.morphology import binary_opening, binary_erosion, binary_dilation
from skimage.morphology import disk, remove_small_objects
import skimage.morphology as sm
from skimage.segmentation import watershed
from skimage import data
from skimage.filters import rank, gaussian, sobel
from skimage.util import img_as_ubyte
from skimage import data, util
from skimage.measure import regionprops_table
from skimage.measure import perimeter
from skimage import measure
from skimage.exposure import rescale_intensity, histogram
from skimage.feature import peak_local_max
import os
import glob
import pandas as pd
from pandas import DataFrame
from scipy.ndimage.morphology import binary_fill_holes
#from skimage.viewer import ImageViewer
from skimage import img_as_float
import time
import base64
from datetime import datetime
import xml.etree.ElementTree as xml
from random import randint

def display_image_label(ch,mask,lable_draw, font_select,font_size):
    '''
    ch: 16 bit input image
    mask: mask for labale
    lable_draw: 'label' or 'area'
    font_select: copy font to the working directory ("DejaVuSans.ttf" eg)
    font_size: 4 is nice size

    return:
    info_table: table of objects measure
    PIL_image: 16 bit mask rgb of the labeled image
    '''
    # count number of objects in nuc['sort_mask']
    info_table = pd.DataFrame(
        measure.regionprops_table(
            mask,
            intensity_image=ch,
            properties=['area', 'label', 'centroid'],
        )).set_index('label')
    PIL_image = Image.fromarray(np.uint16(mask)).convert('RGB')
    #Label ROI by centroid and index
    info_table['label'] = range(2, len(info_table) + 2)
    #round
    info_table = info_table.round({'centroid-0': 0, 'centroid-1': 0})
    info_table = info_table.reset_index(drop=True)
    draw = ImageDraw.Draw(PIL_image)
    # use a bitmap font
    font = ImageFont.truetype(font_select, font_size)
    if lable_draw=='label':
        sel_lable=3
    else:
        sel_lable = 0
    for i in range(len(info_table)):
        draw.text((info_table.iloc[i, 2].astype('int64'), info_table.iloc[i, 1].astype('int64')),str(info_table.iloc[i, sel_lable]), 'red', font=font)
    return info_table, PIL_image

def outline_seg(mask,index):
    seg_mask_temp = np.zeros(np.shape(mask), dtype=np.int32)
    seg_mask_temp[mask == index] = index
    seg_mask_eros_9 = binary_erosion(seg_mask_temp, structure=np.ones((9, 9))).astype(np.float64)
    seg_mask_eros_3 = binary_erosion(seg_mask_temp, structure=np.ones((3, 3))).astype(np.float64)
    framed_mask = seg_mask_eros_3 - seg_mask_eros_9
    return framed_mask

def sum_seg(dict,mask):
    update_mask = np.zeros(np.shape(mask), dtype=np.int32)

    for item in update_mask.items():
        update_mask = update_mask + item
    return update_mask


def binary_frame_mask(ch,mask,table=None):
    '''
    Create a mask for NIS-elements to photo-activate for multiple point
    :parameter
    ch - input image
    mask - input mask (RGB) 32integer
    :return
    framed_mask (RGB)
    '''
    if table is None:
        info_table = pd.DataFrame(
            measure.regionprops_table(
                mask,
                intensity_image=ch,
                properties=['area', 'label', 'centroid'],
            )).set_index('label')
    else:
        info_table=table
    info_table['label'] = range(2, len(info_table) + 2)
    framed_mask = np.zeros(np.shape(mask), dtype=np.int32)
    if len(info_table['label']) > 1:
        for i in info_table.index.values:
            framed_mask = framed_mask + outline_seg(mask,i)
    else:
        framed_mask = outline_seg(mask,info_table.index.values)
    return framed_mask

def binary_frame_mask_single_point(mask,table=None):
    '''
    Create a mask for NIS-elements to photo-activate for single point
    :parameter
    mask - input mask (RGB)
    :return
    framed_mask (RGB)
    '''
    seg_mask_eros_9 = binary_erosion(mask, structure=np.ones((9, 9))).astype(np.float64)
    seg_mask_eros_3 = binary_erosion(mask, structure=np.ones((3, 3))).astype(np.float64)
    seg_frame = np.where(seg_mask_eros_9 + seg_mask_eros_3 == 2, 3, seg_mask_eros_3)
    framed_mask = np.where(seg_frame == 3, 0, seg_mask_eros_3)
    return framed_mask




def Centroid_map(ch, mask, mat):
    '''
        Returns center of mask map after dilation
        :parameter
        ch - Grayscale input image
        mask -
        mat - matrix for the dilation operation
        :return
        table - centroid table
        center_map -
    '''
    table = pd.DataFrame(measure.regionprops_table(mask,
                            intensity_image= ch,
                            properties=['label', 'centroid_weighted',])
                         ).set_index('label')
    center_map = np.zeros(np.shape(mask))
    y = table
    centroid_0 = np.array(y['centroid_weighted-0']).astype('int')
    centroid_1 = np.array(y['centroid_weighted-1']).astype('int')
    center_map[centroid_0, centroid_1] = 1
    center_map = binary_dilation(center_map, structure=np.ones((mat, mat))).astype(np.float64)
    return table, center_map


def seq(start, end, by=None, length_out=None):
    len_provided = True if (length_out is not None) else False
    by_provided = True if (by is not None) else False
    if (not by_provided) & (not len_provided):
        raise ValueError('At least by or n_points must be provided')
    width = end - start
    eps = pow(10.0, -14)
    if by_provided:
        if (abs(by) < eps):
            raise ValueError('by must be non-zero.')
        # Switch direction in case in start and end seems to have been switched (use sign of by to decide this behaviour)
        if start > end and by > 0:
            e = start
            start = end
            end = e
        elif start < end and by < 0:
            e = end
            end = start
            start = e
        absby = abs(by)
        if absby - width < eps:
            length_out = int(width / absby)
        else:
            # by is too great, we assume by is actually length_out
            length_out = int(by)
            by = width / (by - 1)
    else:
        length_out = int(length_out)
        by = width / (length_out - 1)
    out = [float(start)] * length_out
    for i in range(1, length_out):
        out[i] += by * i
    if abs(start + by * length_out - end) < eps:
        out.append(end)
    return out

def evaluate_image_output(mask):
    mask_eval = mask - 1
    mask_eval = np.where(mask_eval == 1, 0, mask_eval)
    if np.sum(mask_eval) < 3:
        mask = np.zeros(np.shape(mask_eval))
    else:
        mask = mask
    return mask

def test_image(arr):
    '''
    test whether the mask generated is empty
    :parameter
    arr - np array
    '''
    values, counts = np.unique(arr.ravel(), axis=0, return_counts=True)
    return counts

def unique_rand(inicial, limit, total):
    '''
    :parameter
    :return
    :Example
    data = unique_rand(1, 60, 6)
    print(data)
    #######################################################
            prints something like
            [34, 45, 2, 36, 25, 32]
    #######################################################
    '''
    data = []
    i = 0
    while i < total:
        number = randint(inicial, limit)
        if number not in data:
            data.append(number)
            i += 1
    return data


import numpy as np
import pandas as pd
from skimage import io, data, exposure, img_as_uint, img_as_ubyte
from PIL import Image
import dash
from dash.dependencies import Input, Output, State
import dash_html_components as html
import dash_core_components as dcc
import dash_table
from dash.exceptions import PreventUpdate
import json
import dash_canvas
from dash_canvas.utils import (image_string_to_PILImage, array_to_data_url)


def parse_jsonstring_line(string):
    """
    Return geometry of line objects.

    Parameters
    ----------

    data : str
        JSON string of data

    """
    try:
        data = json.loads(string)
    except:
        return None
    scale = 1
    props = []
    for obj in data['objects']:
        if obj['type'] == 'image':
            scale = obj['scaleX']
        elif obj['type'] == 'line':
            length = np.sqrt(obj['width']**2 + obj['height']**2)
            scale_factor = obj['scaleX'] / scale
            props.append([scale_factor * length,
                          scale_factor * obj['width'],
                          scale_factor * obj['height'],
                          scale_factor * obj['left'],
                          scale_factor * obj['top']])
    return (np.array(props)).astype(int)


def measure_length(filename):
    '''
    file name is string of the path for example file
    '''
    try:
        img = io.imread(filename)
        # input_gs_image = img
        # input_gs_image = (input_gs_image / input_gs_image.max()) * 255
        img_eq = exposure.equalize_hist(img)
        img = img_as_ubyte(img_eq)
    except:
        img = data.coins()
    height, width = img.shape
    canvas_width = 500
    canvas_height = round(height * canvas_width / width)
    scale = canvas_width / width

    list_columns = ['length', 'width', 'height']
    columns = [{"name": i, "id": i} for i in list_columns]

    app = dash.Dash(__name__)
    server = app.server
    app.config.suppress_callback_exceptions = True


    app.layout = html.Div([
        html.Div([
            dash_canvas.DashCanvas(
                id='canvas-line',
                width=canvas_width,
                height=canvas_height,
                scale=scale,
                lineWidth=2,
                lineColor='red',
                tool="line",
                hide_buttons=['pencil'],
                image_content=array_to_data_url(img),
                goButtonTitle='Measure',
                ),
        ], className="seven columns"),
        html.Div([
        html.Img(src='https://github.githubassets.com/images/modules/logos_page/GitHub-Mark.png', width='30px'),
        html.A(
            id='gh-link',
            children=[
            'View on GitHub'
            ],
            href="http://github.com/plotly/canvas-portal/"
                              "blob/master/apps/measure-length/app.py",
            style={'color': 'black',
                'border':'solid 1px black',
                'float':'left'}
                ),

            html.H3('Draw lines and measure lengths'),
            html.H3(children='How to use this app', id='measure-subtitle'),
            html.Img(id='measure-help',
                     src='assets/measure.gif',
                     width='100%'),
            html.H4(children="Objects properties"),
            dash_table.DataTable(
                id='table-line',
                columns=columns,
                editable=True,
                ),
        ], className="four columns"),
        ])



    @app.callback(Output('canvas-line', 'tool'),
                    [Input('canvas-line', 'image_content')])
    def modify_tool(string):
        return "line"


    @app.callback(Output('measure-subtitle', 'children'),
                    [Input('canvas-line', 'json_data')])
    def reduce_help(json_data):
        if json_data:
            return ''
        else:
            return dash.no_update 


    @app.callback(Output('measure-help', 'width'),
                    [Input('canvas-line', 'json_data')])
    def reduce_help(json_data):
        if json_data:
            return '0%'
        else:
            return dash.no_update 


    @app.callback(Output('table-line', 'data'),
                    [Input('canvas-line', 'json_data')])
    def show_table(string):
        props = parse_jsonstring_line(string)
        df = pd.DataFrame(props[:, :3], columns=list_columns)
        return df.to_dict("records")
    return app





import os
import shutil
import h5py
from pathlib import Path
import numpy as np
import argparse
import pdb

# Assuming the 'default_parameters.h5' file is located directly inside the package folder
#default_parameters_path = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),'default_parameters.h5')
default_parameters_path = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),'parameters.h5')
# Determine the user's home configuration path
user_parameters_path = os.path.join(Path.home(), '.AIPyS', 'parameters.h5') # save the parameter file to .AIPyS folder. 

def ensure_user_parameters():
    """Ensure the user has their own copy of parameters.h5; if not, create it."""
    if not os.path.exists(user_parameters_path):
        os.makedirs(os.path.dirname(user_parameters_path), exist_ok=True)
        shutil.copy(default_parameters_path, user_parameters_path)
        print(f'Created a personal parameters file at {user_parameters_path}')


def update_auto_parameters(parameter_updates,user_parameters_path):
    '''
    this function is used for updating h5 file based on results
    e.g. intercept and slope from model build
    '''
    with h5py.File(user_parameters_path, 'r+') as hdf:
        for key, value in parameter_updates.items():
            if key in hdf:
                value = 'None' if value is None else value
                if isinstance(value, str):
                    hdf[key][()] = np.array(value, dtype=h5py.special_dtype(vlen=str))
                else:
                    hdf[key][...] = value
            else:
                # If the key does not exist in the HDF5 file, create a new dataset.
                if isinstance(value, str):
                    dt = h5py.special_dtype(vlen=str)
                    hdf.create_dataset(key, data=np.array(value, dtype=dt), dtype=dt)
                else:
                    hdf.create_dataset(key, data=value)
                print(f"'{key}' added to or updated in the HDF5 file with value: {value}")



def update_user_parameters(parameter_updates):
    """Example function to update parameters in the user's parameters.h5 file."""
    ensure_user_parameters()  # Ensure the user copy exists
    #pdb.set_trace()
    with h5py.File(user_parameters_path, 'r+') as hdf:
        for key, value in parameter_updates.items():
            if key in hdf:
                value = 'None' if value is None else value
                if isinstance(value, str):
                    hdf[key][()] = np.array(value, dtype=h5py.special_dtype(vlen=str))
                else:
                    hdf[key][...] = value
            else:
                # If the key does not exist in the HDF5 file, create a new dataset.
                if isinstance(value, str):
                    dt = h5py.special_dtype(vlen=str)
                    hdf.create_dataset(key, data=np.array(value, dtype=dt), dtype=dt)
                else:
                    hdf.create_dataset(key, data=value)
                print(f"'{key}' added to or updated in the HDF5 file with value: {value}")

def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error("The file %s does not exist!" % arg)
    else:
        return open(arg, 'r')  # return an open file handle

# def read_user_parameters():
#     """Example function to read parameters from the user's parameters.h5 file."""
#     ensure_user_parameters()  # Ensure the user copy exists
#     with h5py.File(user_parameters_path, 'r') as hdf:
#         # Read and return parameters as needed

def main():
    parser = argparse.ArgumentParser(description='Update or add parameters and display the content of the parameters.h5 file. This script allows users to manage and view parameters stored in a .h5 file. Users can update parameter values or simply display the current configuration.', formatter_class=argparse.RawTextHelpFormatter)
    # Adding detailed help messages for each argument
    parser.add_argument('--data_dir', type=str, help="Directory where data is stored.\nThis path is used to locate your data files for processing.")
    parser.add_argument('--files', type=str, help="Path to the images or files.\nSpecify a single file or a directory containing multiple files.")
    parser.add_argument('--videoName', type=str, help="Name of the output video file.\nDetermine the filename for the processed output video.")
    parser.add_argument('--Image_name', type=str, help="Name of the image file.\nUsed to specify a particular image file name for operations.")
    parser.add_argument('--outPath', type=str, help="Output directory path.\nDefines where to store output files.")
    parser.add_argument('--block_size_cyto', type=int, help="Block size for cytoplasm segmentation.\nAn integer value affecting segmentation precision.")
    parser.add_argument('--offset_cyto', type=int, help="Offset applied to block size in cytoplasm segmentation.\nAdjusts the segmentation algorithm's sensitivity.")
    parser.add_argument('--global_ther', type=float, help="Global threshold value.\nUsed in various thresholding operations; a float between 0 and 1.")
    parser.add_argument('--clean', type=int, help="Parameter for cleaning the image.\nSpecify the intensity of the cleaning operation; higher values for more aggressive cleaning.")
    parser.add_argument('--channel', type=int, help="Channel of the image to be processed.\nIndicates which color channel to use in multi-channel image files.")
    parser.add_argument('--bit8', type=int, help="Bit depth to be considered.\nChoose between 8, 16, or 32-bit processing.")
    parser.add_argument('--ci', type=int, help="Color intensity.\nAdjusts the intensity level for color processing.")
    parser.add_argument('--start_kernel', type=int, help="Initial kernel size for operations that use a kernel.\nDetermines the starting size of the kernel in algorithms that dynamically adjust it.")
    parser.add_argument('--end_karnel', type=int, help="Final kernel size for adaptive algorithms.\nSets the maximum allowable kernel size.")
    parser.add_argument('--kernel_size', type=int, help="Kernel size for processing.\nSpecifies the fixed size of the kernel used in operations like filtering.")
    parser.add_argument('--outputImageSize', type=int, help="Output image size in pixels.\nControls the dimensions of the output image file.")
    parser.add_argument('--windowSize', type=int, help="Size of the window for image processing.\nDefines the window size for operations that process the image in sections.")
    parser.add_argument('--rnadomWindows', type=int, help="Number of random windows to be considered in processing.\nSpecifies how many random windows the algorithm should use.")
    parser.add_argument('--kernelGran', type=int, help="Granularity of the kernel.\nFine-tunes the granularity for kernel-based operations.")
    parser.add_argument('--w', type=int, help="Width of the image in pixels.\nUsed to specify the target width for image resizing operations.")
    parser.add_argument('--h', type=int, help="Height of the image in pixels.\nSets the target height for image adjustments.")
    parser.add_argument('--diameter', type=int, help="cellpose parameter.\nEstimation of cell diameter.")
    parser.add_argument('--model_type',type=str, help=" <cyto> or model_type=<nuclei>")
    parser.add_argument('--channels',type=str,help = "grayscale cellpose")
    parser.add_argument('--extract_pixel',type=int,help = "size of extraction according to mask (e.g. 50 pixel)")
    parser.add_argument('--resize_pixel',type=int,help = "resize for preforming tf prediction (e.g. 150 pixel)")
    parser.add_argument('--dataPath',type=str,help = "\imageSequence\data")
    parser.add_argument('--imagePath',type=str,help = "\imageSequence\images")
    parser.add_argument('--imW',type=int,help = "10")
    parser.add_argument('--imH',type=int,help = "5")
    parser.add_argument('--thold',type=float,help = "0.5")
    parser.add_argument('--imagesN',type=int,help = "number of images for analyse")
    parser.add_argument('--trainingDataPath',type=str,help = "use all the images in the base_training_directory")
    parser.add_argument('--areaSel',type=int,help = "area cutoff")
    parser.add_argument('--fractionData',type=int,help = "use fraction of the data for training")
    parser.add_argument('--intercept',type=float,help = "intercept estimated by the bayesian model")
    parser.add_argument('--slope',type=float,help = "intercept estimated by the bayesian model")
    
    # Argument for displaying the parameters without making updates
    parser.add_argument('--display', action='store_true', help="Display the contents of the parameters file without making any updates.\nUse this option to view the current settings stored in the .h5 file.")

    
    args = parser.parse_args()
   
    # Construct the dictionary from provided CLI arguments
    # Exclude 'h5_file' from update_dict, as it's not a parameter to store but the file path
    update_dict = {k: v for k, v in vars(args).items() if k != 'h5_file' and v is not None}
    #pdb.set_trace()
    # Update or add parameters in the H5 file
    update_user_parameters(parameter_updates = update_dict)


if __name__ == "__main__":
    main()




from AIPyS.classification.bayes.RunningWindow import RunningWindow
from AIPyS.supportFunctions.GranularityFunc import openingOperation,resize,imageToRGB
from AIPyS.supportFunctions.AIPS_file_display import Compsite_display
from AIPyS.segmentation.cellpose.StackObjects_cellpose import StackObjects_cellpose 
from AIPyS.supportFunctions.AIPS_functions import areaThreshold
import string
import cv2
import pdb
import random
import numpy as np
import skimage
import os
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from PIL import Image, ImageEnhance, ImageDraw,ImageFont
from IPython.display import clear_output

class GranularityDataGen(RunningWindow):
    def __init__(self,kernelGran,w, h ,*args, **kwargs):
        # require self.outPath(DataLoad) 
        self.kernelGran = kernelGran # The kernelGran is dictated by the GranularityVideo generated video file of Granularity process (example can be 6)
        self.w, self.h =w,h # for resizing
        super().__init__(*args, **kwargs)

    def addText_splitVideo(self,image_input,text):
        return cv2.putText(image_input, f'Imagename: {text[0]} \n Ratio: {text[1]} ', (10,50), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1, cv2.LINE_AA) #0.5 font, 1 thickness
    
    
    def generate_random_string(self,length):
        letters = string.ascii_letters  # this includes both lowercase and uppercase letters
        result_str = ''.join(random.choice(letters) for i in range(length))
        return result_str 

    def imageMeasure(self,inage_input, mask):
        return np.mean(inage_input[mask>0]), np.std(inage_input[mask>0])
      
    def getMeanSD(self,input_image):
        return np.round(np.mean(input_image),3),np.round(np.std(input_image),3)

    def granCalc(self):
        '''
        Produce output table used for Bayes training (without the label option)
        and Video file with Ratio value and image file name.
        Ratio is the granular spectrum
        '''
        # pdb.set_trace()
        image_name = self.generate_random_string(7)
        tempFolder = os.path.join(self.outPath,image_name)
        ImageOrig = os.path.join(tempFolder,'ImageOrig')
        ImageSplit = os.path.join(tempFolder,'ImageSplit')
        os.makedirs(tempFolder,exist_ok=True)
        os.makedirs(ImageOrig,exist_ok=True)
        os.makedirs(ImageSplit,exist_ok=True)
        image = self.img # should be single image
        mask = self.cytosolSegmentation(image)
        # creat original image:
        mean,sd = self.imageMeasure(image, mask)
        skimage.io.imsave(os.path.join(ImageOrig,image_name + f'_mean_{mean}' + f'_sd_{sd}' '.tif'), image)
        imageSlice,maskSlice = self.runningWin(image_input = image, mask = mask)
        if imageSlice is None or maskSlice is None:
            raise ValueError('None is not a valid list')
        dfGran = {'name':[],'ratio':[],'intensity':[],'sd':[]}
        
        open_vec = np.linspace(2,self.kernelGran, self.kernelGran, endpoint=True, dtype=int).tolist()
        open_vec = list(set([open for open in open_vec if open % 2 ==0 ]))
        # video gen
        open_vec = np.linspace(2, self.kernelGran, self.kernelGran, endpoint=True, dtype=int)
        frame = self.disply_compsite(imageSlice[0], maskSlice[0])
        width,height = self.w,self.h
        video = cv2.VideoWriter(os.path.join(ImageSplit,image_name + '.avi'), cv2.VideoWriter_fourcc(*'MJPG'),1.0,(width,height),  isColor = True)
        for i,(imageCrop,maskCrop) in enumerate(zip(imageSlice,maskSlice)):
            prev_image = imageCrop
            for open in open_vec:
                imageDecy = openingOperation(kernel = open, image = prev_image)
                if np.sum(imageDecy[maskCrop>0]) == 0:
                    denominator = np.sum(prev_image[maskCrop > 0]) + 0.000001
                else:
                    denominator = np.sum(prev_image[maskCrop > 0])
                ratio = np.sum(prev_image[maskCrop > 0] - imageDecy[maskCrop > 0])/denominator
                prev_image = imageDecy
            mean,sd = self.imageMeasure(imageDecy, maskCrop)
            dfGran['name'].append(image_name + f'_i')
            dfGran['ratio'].append(ratio)
            dfGran['intensity'].append(mean)
            dfGran['sd'].append(sd)
            frame = self.disply_compsite(imageCrop, maskCrop)
            frame = resize(frame, width ,height)
            print('-' * i, end = '\r')
            self.addText_splitVideo(frame,(image_name + f'_{i}',np.round(ratio,4)))
            video.write(frame)
        video.release()
        dfGran = pd.DataFrame(dfGran)
        dfGran['label'] = 0
        dfGran.to_csv(os.path.join(ImageSplit,image_name + '.csv'))
        
    def displayImageFrame(self,add_text = True):
        '''
        return RGB image with running windows frames and coordinates mounted
        '''
        image = self.img # should be single image
        mask = self.cytosolSegmentation(image)
        imgFrame = np.zeros_like(image,dtype='int8') # mask for the frame size
        coordinate = {'x':[],'y':[]}
        if not isinstance(self.windowSize, (int, float)):
                raise ValueError("window size is missing")
        if not isinstance(self.rnadomWindows, (int, float)):
            raise ValueError("rando crop generator is missing choose a number")
        minAxsis = min(image.shape[0],image.shape[1])
        winSize = self.windowSize
        while minAxsis % winSize != 0:
            minAxsis -= 1
        stepSize =  int(minAxsis/winSize)
        print(f'max image size: {minAxsis}, step size: {stepSize}')
        i = 0
        for y in range(0,minAxsis,stepSize):
            for x in range(0,minAxsis,stepSize):
                imgFrame[y:y+1, x:x+stepSize] = 255
                # Draw bottom border of the rectangle
                imgFrame[y+stepSize-1:y+stepSize, x:x+stepSize] = 255
                # Draw left border of the rectangle
                imgFrame[y:y+stepSize, x:x+1] = 255
                # Draw right border of the rectangle
                imgFrame[y:y+stepSize, x+stepSize-1:x+stepSize] = 255
                coordinate['x'].append(y)
                coordinate['y'].append(x)
                print('-' * i, end = '\r')
                i+=1
        rnadomWindows = [1]*self.rnadomWindows
        range_xy = 0+stepSize,minAxsis-stepSize
        while rnadomWindows:
            x = np.random.randint(low = range_xy[0],high = range_xy[1],size =1).tolist()[0]
            imgFrame[x,x:x+stepSize] = 255
            imgFrame[x+stepSize,x:x+stepSize] = 255
            imgFrame[x:x+stepSize,x] = 255
            imgFrame[x:x+stepSize,x+stepSize] = 255
            coordinate['x'].append(x)
            coordinate['y'].append(x)
            rnadomWindows.pop()
        imgFrameMount = imageToRGB(image = image, mask = imgFrame)
        with PdfPages(os.path.join(self.outPath,'imageFrame.pdf')) as pdf:
            plt.rcParams['figure.dpi'] = 150
            plt.rcParams['font.family'] = ['serif']
            plt.rcParams['font.size'] = 12
            plt.rcParams['axes.labelsize'] = 12
            plt.rcParams['axes.titlesize'] = 12
            plt.rcParams['xtick.labelsize'] = 12
            plt.rcParams['ytick.labelsize'] = 12
            if add_text:
                corr_list = [i for i in zip(coordinate["x"],coordinate["y"])]
                plt.figure(figsize=(10, 10))
                for i in corr_list:
                    cv2.putText(imgFrameMount, f'{i}',i , cv2.FONT_HERSHEY_SIMPLEX, 1, (255,255,255), 1, cv2.LINE_AA)
                plt.imshow(imgFrameMount)
                pdf.savefig()
                plt.close()
            else:
                plt.figure(figsize=(10, 10))
                plt.imshow(imgFrameMount)
                pdf.savefig()
                plt.close()
        
            
    def check_list(self):
        if not isinstance(self.Image_name, list):
            raise ValueError("Requries list of images path")
    
    def granCalc_imageGen(self):
        '''
        Generates image per slice and a single table. for training
        Note - expects a list of images path
        Produce output table used for Bayes training (without the label option)
        and Video file with Ratio value and image file name.
        Ratio is the granular spectrum
        '''
        # pdb.set_trace()
        self.check_list()
        image_name_list = self.Image_name
        tempFolder = os.path.join(self.outPath,'imageSequence')
        ImagePath = os.path.join(tempFolder,'images')
        ImageOrig = os.path.join(tempFolder,'ImageOrig')
        DataPath = os.path.join(tempFolder,'data')
        os.makedirs(tempFolder, exist_ok=True)
        os.makedirs(ImagePath, exist_ok=True)
        os.makedirs(ImageOrig,exist_ok=True)
        os.makedirs(DataPath, exist_ok=True)
        dfGran = {'name':[],'ratio':[],'intensity':[],'sd':[],'maskArea':[]}
        dfGranImageTotal = {'name':[],'intensity':[],'sd':[],'maskArea':[]}
        for image_path in image_name_list:
            image = skimage.io.imread(image_path)
            image_name = self.generate_random_string(7)
            mask = self.cytosolSegmentation(image)
            # creat original image:
            mean,sd = self.imageMeasure(image, mask)
            maskArea = np.sum(mask>0)/np.sum(np.ones_like(mask))
            #image intensity and sd before slicing:
            dfGranImageTotal['name'].append(image_name + f'.png')
            dfGranImageTotal['intensity'].append(mean)
            dfGranImageTotal['sd'].append(sd)
            dfGranImageTotal['maskArea'].append(maskArea)
            skimage.io.imsave(os.path.join(ImageOrig,image_name + f'_mean_{mean}' + f'_sd_{sd}' '.tif'), image)
            imageSlice,maskSlice = self.runningWin(image_input = image, mask = mask)
            if imageSlice is None or maskSlice is None:
                raise ValueError('None is not a valid list')
            open_vec = np.linspace(2,self.kernelGran, self.kernelGran, endpoint=True, dtype=int).tolist()
            open_vec = list(set([open for open in open_vec if open % 2 ==0 ]))
            # video gen
            open_vec = np.linspace(2, self.kernelGran, self.kernelGran, endpoint=True, dtype=int)
            clear_output(wait=True)
            for i,(imageCrop,maskCrop) in enumerate(zip(imageSlice,maskSlice)):
                clear_output(wait=True)
                prev_image = imageCrop
                maskArea = np.sum(maskCrop>0)/np.sum(np.ones_like(maskCrop))
                if maskArea < 0.4:
                    continue
                else:
                    for open in open_vec:
                        imageDecy = openingOperation(kernel = open, image = prev_image)
                        if np.sum(imageDecy[maskCrop>0]) == 0:
                            denominator = np.sum(prev_image[maskCrop > 0]) + 0.000001
                        else:
                            denominator = np.sum(prev_image[maskCrop > 0])
                        ratio = np.sum(prev_image[maskCrop > 0] - imageDecy[maskCrop > 0])/denominator
                        prev_image = imageDecy
                    mean,sd = self.imageMeasure(imageDecy, maskCrop)
                    dfGran['name'].append(image_name + f'_{i}.png')
                    dfGran['ratio'].append(ratio)
                    dfGran['intensity'].append(mean)
                    dfGran['sd'].append(sd)
                    dfGran['maskArea'].append(maskArea)
                    frame = self.disply_compsite(imageCrop, maskCrop)
                    frame = resize(frame, 500 ,500)
                    PIL_image_grey = Image.fromarray(frame).convert('L') # convery to gray scale
                    PIL_image = PIL_image_grey.convert('RGB') # for getting a grey scale with red text
                    draw = ImageDraw.Draw(PIL_image)
                    font_size = 24
                    font = ImageFont.truetype("arial.ttf", font_size)  # Adjust as necessary
                    draw.text((5, 5),f'Imagename:{image_name}_{i} \n Ratio:{np.round(ratio,2)} \n Area: {np.round(maskArea,2)}', 'red',font=font)
                    print('-' * i, end = '\r')
                    PIL_image.save(os.path.join(ImagePath,image_name + '_' + f'{i}.png'))
        dfGran = pd.DataFrame(dfGran)
        dfGran['label'] = 0
        dfGran.to_csv(os.path.join(DataPath,'imageseq_data.csv'))
        dfGranImageTotal = pd.DataFrame(dfGranImageTotal)
        dfGranImageTotal.to_csv(os.path.join(DataPath,'imageOrig_data.csv'))
        
    

class GranularityDataGen_cp(StackObjects_cellpose):
    def __init__(self,trainingDataPath,kernelGran,w, h ,*args, **kwargs):
        # require self.outPath(DataLoad) 
        self.trainingDataPath = trainingDataPath
        self.kernelGran = kernelGran # The kernelGran is dictated by the GranularityVideo generated video file of Granularity process (example can be 6)
        self.w, self.h =w,h # for resizing
        super().__init__(*args, **kwargs)
        self.granCalc_imageGen_cp()
        
    def addText_splitVideo(self,image_input,text):
        return cv2.putText(image_input, f'Imagename: {text[0]} \n Ratio: {text[1]} ', (10,50), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1, cv2.LINE_AA) #0.5 font, 1 thickness
    
    
    def generate_random_string(self,length):
        letters = string.ascii_letters  # this includes both lowercase and uppercase letters
        result_str = ''.join(random.choice(letters) for i in range(length))
        return result_str 

    def imageMeasure(self,image_input):
        pass
      
    def getMeanSD(self,image_input):
        return np.round(np.mean(image_input),3),np.round(np.std(image_input),3)
    
    def check_list(self):
        if not isinstance(self.Image_name, list):
            raise ValueError("Requries list of images path")
    
    def sdCalc(self,image,mean_value):
        squared_diff_sum = 0
        num_elements = 0
        for row in image:
            for element in row:
                squared_diff_sum += (element - mean_value) ** 2
                num_elements += 1
        return (squared_diff_sum / num_elements)**0.5
        
    
    def granCalc_imageGen_cp(self):
        '''
        Adjusted to cellpose, Generates image per slice and a single table. for training
        Note - expects a list of images path
        Produce output table used for Bayes training (without the label option)
        and Video file with Ratio value and image file name.
        Ratio is the granular spectrum
        '''
        # pdb.set_trace()
        self.check_list()
        image_name_list = self.Image_name
        tempFolder = os.path.join(self.outPath,'imageSequence')
        ImagePath = os.path.join(tempFolder,'images')
        DataPath = os.path.join(tempFolder,'data')
        os.makedirs(tempFolder, exist_ok=True)
        os.makedirs(ImagePath, exist_ok=True)
        os.makedirs(DataPath, exist_ok=True)
        dfGran = {'name':[],'ratio':[],'intensity':[],'sd':[],'maskArea':[]}
        c = 0
        for image_path in image_name_list:
            image = skimage.io.imread(image_path)
            image_name = self.generate_random_string(7)
            # Initiate cellpose segmentation
            mask, table = self.cellpose_segmentation(image_input = image)
            # add MLE cleaning area step if area is smaller then th skip it:
            th_area = areaThreshold(arr_area = table.area.values)
            if th_area=="na":
                #incase MLE fail
                th_area = 1
            # create original image:
            #print('-' * c, end = '\r')
            clear_output(wait=True)
            for idx, (index, row) in enumerate(table.iterrows()):
                if idx % 10 == 0:
                    c += 1
                    print('.'*c, end = '\r')
                img_label = row['label'] # first cell analysis 
                try:
                    stack_img, stack_mask = self.stackObjects_cellpose_ebimage_parametrs_method(image,mask,table,img_label)
                except:
                    continue
                if stack_img is None:
                    continue
                intensity =  np.mean(stack_img)
                maskArea = row['area']
                if maskArea < th_area:
                    continue
                sd = self.sdCalc(stack_img,intensity)
                #image intensity and sd before opening:
                # # video gen
                imageDecy = openingOperation(kernel = self.kernelGran, image = stack_img)
                ratio = np.mean(imageDecy)/intensity
                dfGran['name'].append(image_name + f'_{idx}.png')
                dfGran['ratio'].append(ratio)
                dfGran['intensity'].append(intensity)
                dfGran['sd'].append(sd)
                dfGran['maskArea'].append(maskArea)
                frame_obj = Compsite_display(stack_img, stack_mask, 2)
                frame = resize(image=frame_obj.rgb_out, width=self.w, hight=self.h)
                PIL_image_grey = Image.fromarray(frame).convert('L') # convery to gray scale
                PIL_image = PIL_image_grey.convert('RGB') # for getting a grey scale with red text
                draw = ImageDraw.Draw(PIL_image)
                font_size = 24
                font = ImageFont.truetype("arial.ttf", font_size)  # Adjust as necessary
                draw.text((5, 5),f'Imagename:{image_name}_{idx} \n Ratio:{np.round(ratio,4)} \n Area: {np.round(maskArea,4)}', 'red',font=font)
                PIL_image.save(os.path.join(ImagePath,image_name + '_' + f'{idx}.png'))
        dfGran = pd.DataFrame(dfGran)
        dfGran['label'] = "na"
        dfGran.to_csv(os.path.join(DataPath,'imageseq_data.csv'))
            
    


from AIPyS.segmentation.parametric.GlobalSeg import GlobalSeg
import os
import numpy as np
import cv2 
from IPython.display import clear_output
import pdb
from PIL import Image, ImageDraw,ImageFont

class ImageSeqGlobSeg(GlobalSeg):
    '''
    Image sequences of segmented files are generated and saved as global segmented files, stored in the MP4 format. 
    These are expected to be 8-bit images with the shape (w, h, c), where 'c' represents 3 channels.
    '''
    # def __init__(self,videoName,Image_name,outPath,block_size_cyto, offset_cyto, global_ther, clean, channel, bit8, ci):
    #     self.videoName = videoName #.mp4
    #     super().__init__(Image_name,outPath,block_size_cyto, offset_cyto, global_ther, clean, channel, bit8, ci)

    def __init__(self,videoName,*args, **kwargs):
        self.videoName = videoName #.mp4
        super().__init__(*args, **kwargs)
        
    def check_multiImage(self):
        if len(self.img) < 2:
            assert "list of images are required"

    # def rgbMasking(self,image):
    #     mask = self.cytosolSegmentation(image = image)
    #     frame = self.disply_compsite(image = image, mask = mask)
    #     return frame,mask
    
    def addText(self,image_input,text):
        return cv2.putText(image_input, f'mean: {text[0]} \n sd: {text[1]}', (10,50), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 2, cv2.LINE_AA)
    
    def resize(self,image):
        resized = cv2.resize(image, (500,500), interpolation = cv2.INTER_AREA)
        return resized
        
    def imageSeqGen(self):
        self.check_multiImage()
        images = self.img
        #pdb.set_trace()
        frame,mask = self.rgbMasking(image = images[0])
        width,height,layers = frame.shape
        frame = self.resize(frame)
        video = cv2.VideoWriter(os.path.join(self.outPath,self.videoName), cv2.VideoWriter_fourcc(*'MJPG'),1.0,(500,500),  isColor = True)
        #clear_output(wait=True)
        for i, image in enumerate(images):
            mean,sd = np.round(np.mean(image),2), np.round(np.std(image),2)
            frame,mask = self.rgbMasking(image = image)
            # add text:
            print('-' * i, end = '\r')
            #pilimage = Image.fromarray(frame)
            frame = self.resize(frame)
            self.addText(frame,(mean,sd))
            video.write(frame)
        video.release()
        


# Measuring Cell Diameter
# python measure_diameter.py --debug
import argparse
import random
import glob
from pathlib import Path
import os

import threading
import webbrowser
import time

from web_app.measure_length.app import measure_length
from web_app.Image_labeling.app import image_labeling
from web_app.TableViz.app import data_viz
from AIPyS.segmentation.cellpose.AIPS_cellpse_video_gen import CellPoseSeg
from AIPyS.classification.bayes.GranulaityMesure import GranulaityMesure_cp
from AIPyS.classification.bayes.GranularityDataGen import GranularityDataGen_cp
from AIPyS.classification.bayes.Baysian_training import Baysian_training
from AIPyS.classification.bayes.Baysian_deploy import BayesDeploy
from AIPyS.supportFunctions.unpack_h5 import parametersInspec
from AIPyS.CLI.set_parameters import update_auto_parameters

# location of parameters file
user_parameters_path = os.path.join(Path.home(), '.AIPyS', 'parameters.h5')

def run(option):
    if option == "measDia":
        parameters = parametersInspec(option,user_parameters_path)
        image_name = parameters["Image_name"]
        app = measure_length(image_name)
        app.run_server()
    elif option == "cp_seg_video":
        parameters = parametersInspec(option,user_parameters_path)
        files = glob.glob(parameters['data_dir'] + "\\*.tif")
        if parameters['channels']=='greyscale':
            channels = [0,0]
        else:
            print('channels are missing')
        CellPoseSeg(diameter =  parameters['diameter'],videoName = parameters['videoName'], model_type = parameters['model_type'], channels = channels,
                    Image_name = files[:parameters['imagesN']], outPath = parameters['outPath'])
    elif option == "cp_gran_video":
        # Placeholder for cp_gran_video option - creates a video of kernel opening
        parameters = parametersInspec(option,user_parameters_path)
        if parameters['channels']=='greyscale':
            channels = [0,0]
        else:
            print('channels are missing')
        GranulaityMesure_cp(start_kernel = parameters['start_kernel'],end_karnel = parameters['end_karnel'],kernel_size = parameters['kernel_size'],
                            extract_pixel = parameters['extract_pixel'],outputImageSize = parameters['outputImageSize'], resize_pixel = parameters['resize_pixel'],
                            videoName = parameters['videoName'], model_type = parameters['model_type'], channels = parameters['channels'],
                            Image_name = parameters['Image_name'], outPath = parameters['outPath'],diameter =  parameters['diameter']) # save video form cell segmentation
    elif option == "cp_gran_table_gen":
        ### trainingDataPath needs to be updated 
        parameters = parametersInspec(option,user_parameters_path)
        if parameters['channels']=='greyscale':
            channels = [0,0]
        else:
            print('channels are missing')
        pattern = os.path.join(parameters['trainingDataPath'], '**', '*.tif')
        # List all the .tif files
        files = glob.glob(pattern, recursive=True)
        random.shuffle(files)
        GranularityDataGen_cp(kernelGran = parameters['kernelGran'],w = parameters['w'],h = parameters['h'],
                            extract_pixel = parameters['extract_pixel'], resize_pixel = parameters['resize_pixel'],diameter =  parameters['diameter'],
                            model_type = parameters['model_type'], channels = channels,
                            Image_name = files[:parameters['imagesN']], outPath = parameters['outPath'],trainingDataPath = parameters['trainingDataPath'])  
        # Placeholder for table generation - ration tables for labeling with corresponding images
    elif option == "dataLabeling":
        # Placeholder for app for labeling images
        parameters = parametersInspec(option,user_parameters_path)
        dataPath = parameters['dataPath']
        imagePath = parameters['imagePath']
        app = image_labeling(imagePath,dataPath)
        app.run_server()
        #port = 8050
        #app.run_server(port = port)
        #webbrowser.open_new(f'http://localhost:{port}')
    elif option == "data_viz":
        # Placeholder for app for data vitalization after labeling
        parameters = parametersInspec(option,user_parameters_path)
        dataPath = parameters['dataPath']
        imagePath = parameters['imagePath']
        app = data_viz(imagePath,dataPath)
        app.run_server()
    elif option == "modelBuild":
        # Placeholder for modelBuild option functionality
        parameters = parametersInspec(option,user_parameters_path)
        if parameters['channels']=='greyscale':
            channels = [0,0]
        else:
            print('channels are missing')
        model = Baysian_training(dataPath = parameters['dataPath'], imW = parameters['imW'], imH = parameters['imH'], thold = parameters['thold'],
                        extract_pixel = parameters['extract_pixel'], resize_pixel = parameters['resize_pixel'],
                         model_type = parameters['model_type'], channels = channels,Image_name = parameters['Image_name'],
                         outPath = parameters['outPath'], diameter =  parameters['diameter'],areaSel = parameters['areaSel'],fractionData = parameters['fractionData']).bayesModelTraining()
        updateParaDict = {"intercept":float(model[0]),"slope":float(model[1])} 
        update_auto_parameters(parameter_updates = updateParaDict, user_parameters_path = user_parameters_path)
        print("Granularity model parameters estimation is done")
    elif option == "deployBuild":
        # Placeholder for deployBuild option functionality
        parameters = parametersInspec(option,user_parameters_path)
        if parameters['channels']=='greyscale':
            channels = [0,0]
        else:
            print('channels are missing')
        parameters = parametersInspec(option,user_parameters_path)
        BayesDeploy(intercept = parameters['intercept'], slope = parameters['slope'], kernelGran = parameters['kernelGran'], dataPath = parameters['dataPath'], imW = parameters['imW'], imH = parameters['imH'], thold = parameters['thold'],
                    extract_pixel = parameters['extract_pixel'], resize_pixel = parameters['resize_pixel'],areaSel = parameters['areaSel'],fractionData = parameters['fractionData'],
                     model_type = parameters['model_type'], channels = channels, Image_name = parameters['Image_name'], outPath = parameters['outPath'], diameter =  parameters['diameter'],)
    else:
        print('Invalid option. Please choose a valid option.')

def main():
    # Create the parser
    parser = argparse.ArgumentParser(description='Application for measuring cell diameter.')
    # Add an argument for the option selection
    parser.add_argument('--option', type=str, # Added the long option '--option' for clarity
                        choices=['measDia', 'cp_seg_video', 'cp_gran_video', 'cp_gran_table_gen', 'dataLabeling', 'data_viz', 'modelBuild', 'deployBuild'],
                        required=True,
                        help='Select an option to run: measDia, cp_seg_video, cp_gran_video, cp_gran_table_gen, dataLabeling, data_viz, modelBuild, deployBuild')
    args = parser.parse_args()

    # Call the run function with the selected option and debug mode
    run(args.option)


if __name__ == '__main__':
    main()




from AIPyS.segmentation.parametric.GlobalSeg import GlobalSeg
from AIPyS.segmentation.parametric.ImageSeqGlobSeg import ImageSeqGlobSeg
from AIPyS.supportFunctions.GranularityFunc import openingOperation,resize
from AIPyS.supportFunctions.AIPS_file_display import Compsite_display
from AIPyS.segmentation.cellpose.StackObjects_cellpose import StackObjects_cellpose
from skimage import measure, morphology
import os
import numpy as np
import cv2
import time
from IPython.display import clear_output
import pdb

class GranulaityMesure(GlobalSeg):
    '''
    Note: get single tif 16bit image only
    # start_kernel = 2,end_karnel = 50, kernel_size = 20,outputTableName = None,outPath = None,
    '''
    def __init__(self,start_kernel,end_karnel, kernel_size,videoName,outputImageSize,*args, **kwargs):
        self.start_kernel = start_kernel
        self.end_karnel = end_karnel
        self.kernel_size = kernel_size
        self.videoName = videoName
        self.outputImageSize = outputImageSize
        super().__init__(*args, **kwargs)
        
        # def resize(self,image):
        #     resized = cv2.resize(image, (self.outputImageSize,self.outputImageSize), interpolation = cv2.INTER_AREA)
        #     return resized
    
    def addText(self,image_input,text):
        return cv2.putText(image_input, f'Kernel: {text[0]} \n ratio: {text[1]} \n sd: {text[2]}', (10,50), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1, cv2.LINE_AA) #0.5 font, 1 thickness
    
    def GranularityVideo(self):
        open_vec = np.linspace(self.start_kernel, self.end_karnel, self.kernel_size, endpoint=True, dtype=int)
        image = self.img
        frame,mask = self.rgbMasking(image = image)
        # width,height,layers = frame.shape
        frame = resize(image = frame, width = self.outputImageSize,hight = self.outputImageSize) # resize(image,width,hight)
        video = cv2.VideoWriter(os.path.join(self.outPath,self.videoName), cv2.VideoWriter_fourcc(*'MJPG'),1.0,(self.outputImageSize,self.outputImageSize),  isColor = True)
        #pdb.set_trace()
        #clear_output(wait=True)
        prev_image = image
        for i,opning in enumerate(open_vec):
            openImage_curr = openingOperation(kernel = opning,image = prev_image)
            if np.sum(prev_image[mask>0]) == 0:
                denominator = np.sum(prev_image[mask>0])+0.000001
            else:
                denominator = np.sum(prev_image[mask>0])
            ratio = (np.sum(prev_image[mask>0]) - np.sum(openImage_curr[mask>0]))/denominator
            prev_image = openImage_curr
            frame = self.disply_compsite(openImage_curr, mask)
            mean = np.round(ratio,2)
            sd = np.round(np.std(openImage_curr[mask>0]),2)
            frame = resize(image = frame, width = self.outputImageSize,hight = self.outputImageSize)
            print('-' * i, end = '\r')
            # add text:
            self.addText(frame,(opning,mean,sd))
            video.write(frame)
        video.release()

class GranulaityMesure_cp(StackObjects_cellpose):
    '''
    cellprofiler adjuste, Note: get single tif 16bit image only
    # start_kernel = 2,end_karnel = 50, kernel_size = 20,outputTableName = None,outPath = None,
    '''
    def __init__(self,videoName, start_kernel,end_karnel, kernel_size,outputImageSize,*args, **kwargs):
        self.videoName = videoName
        self.start_kernel = start_kernel
        self.end_karnel = end_karnel
        self.kernel_size = kernel_size
        self.outputImageSize = outputImageSize
        super().__init__(*args, **kwargs)
        self.GranularityVideo_cp()
        
    def addText(self,image_input,text):
        return cv2.putText(image_input, f'Kernel: {text[0]} \n ratio: {text[1]} \n sd: {text[2]}', (10,50), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1, cv2.LINE_AA) #0.5 font, 1 thickness
    
    def check_singleImage(self):
        if isinstance(self.img, list):
            assert "should be a path to single image"
    
    def GranularityVideo_cp(self):
        '''
        randomaly chose cell id and proform opening operation on it. 
        '''
        open_vec = np.linspace(self.start_kernel, self.end_karnel, self.kernel_size, endpoint=True, dtype=int)
        self.check_singleImage()
        image = self.img # must be single image
        mask, table = self.cellpose_segmentation(image_input = image)
        # single cell randomly selected:
        img_label = table['label'].sample(n=1).iloc[0]
        stack_img, stack_mask = self.stackObjects_cellpose_ebimage_parametrs_method(image,mask,table,img_label)
        frame_obj = Compsite_display(stack_img, stack_mask, 2)
        frame = resize(image=frame_obj.rgb_out, width=self.outputImageSize, hight=self.outputImageSize)  # resize(image,width,hight)
        video = cv2.VideoWriter(os.path.join(self.outPath,self.videoName), cv2.VideoWriter_fourcc(*'MJPG'),1.0,(self.outputImageSize,self.outputImageSize),  isColor = True)
        video.write(frame)
        prev_image = stack_img
        print('analysis might take a few minutes')
        for i, opning in enumerate(open_vec):
            openImage_curr = openingOperation(kernel=opning, image=prev_image)
            if np.sum(prev_image[stack_mask > 0]) == 0:
                denominator = np.sum(prev_image[stack_mask > 0]) + 0.000001
            else:
                denominator = np.sum(prev_image[stack_mask > 0])
            ratio = (np.sum(prev_image[stack_mask > 0]) - np.sum(openImage_curr[stack_mask > 0])) / denominator
            prev_image = openImage_curr
            frame_obj = Compsite_display(openImage_curr, stack_mask, 2)
            frame = resize(image=frame_obj.rgb_out, width=self.outputImageSize, hight=self.outputImageSize)
            mean = np.round(ratio, 4)
            sd = np.round(np.std(openImage_curr[stack_mask > 0]), )
            print('-' * i, end='\r')
            # add text:
            self.addText(frame, (opning, mean, sd))
            video.write(frame)
        video.release()
        print(f'{self.videoName} is saved')    


from AIPyS.segmentation.cellpose.AIPS_cellpose import AIPS_cellpose
import numpy as np
import skimage

class StackObjects_cellpose(AIPS_cellpose):
    '''
    function similar to the EBimage stackObjectsta, return a crop size based on center of measured mask.

    Parameters
    ----------
    extract_pixel: int
        size of extraction according to mask (e.g. 50 pixel)
    resize_pixel: int
        resize for preforming tf prediction (e.g. 150 pixel)
    img_label: int
        the mask value for stack
    Returns
    -------
    center image with out background (img)
    '''
    def __init__(self,extract_pixel, resize_pixel,*args, **kwargs):
        self.extract_pixel = extract_pixel
        self.resize_pixel = resize_pixel
        super().__init__(*args, **kwargs)
        
    
    def checkMissing(self,inputObj):
        if inputObj is None:
            raise "missing cellpose output masks tables"

    def stackObjects_cellpose_ebimage_parametrs_method(self, image_input,mask,table,img_label):
        '''
        img_label: int
            the mask value for stack
        '''
        self.checkMissing(inputObj = mask)
        self.checkMissing(inputObj = table)
        self.checkMissing(inputObj = img_label)
        #x, y = table.loc[table['label']==img_label, ["centroid-0", "centroid-1"]].values.tolist()
        #row, col = disk((int(x), int(y)), 10)
        arr = table.loc[table['label']==img_label, ["centroid-0", "centroid-1"]].values
        # #x, y = table.loc[table['label']==img_label, ["centroid-0", "centroid-1"]].values
        x, y = int(arr[:,0]), int(arr[:,1])
        mask_value = mask[x, y]
        x_start = x - self.extract_pixel
        x_end = x + self.extract_pixel
        y_start = y - self.extract_pixel
        y_end = y + self.extract_pixel
        if x_start < 0 or x_end < 0 or y_start < 0 or y_end < 0:
            stack_img = None
            mask_value = None
            return stack_img, mask_value
        else:
            mask_bin = np.zeros((np.shape(image_input)[0], np.shape(image_input)[1]), np.int32)
            mask_bin[mask == mask_value] = 1
            masked_image = image_input * mask_bin
            stack_img = masked_image[x_start:x_end, y_start:y_end]
            stack_img = skimage.transform.resize(stack_img, (self.resize_pixel, self.resize_pixel), anti_aliasing=False)
            return stack_img, mask_value


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 31 14:01:39 2021

@author: kanferg
"""
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 31 20:05:23 2017
@author: DIP
@Copyright: Dipanjan Sarkar
"""

from sklearn import metrics
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.preprocessing import LabelEncoder
from sklearn.base import clone
from sklearn.preprocessing import label_binarize
from scipy import interp
from sklearn.metrics import roc_curve, auc 


def get_metrics(true_labels, predicted_labels):
    
    print('Accuracy:', np.round(
                        metrics.accuracy_score(true_labels, 
                                               predicted_labels),
                        4))
    print('Precision:', np.round(
                        metrics.precision_score(true_labels, 
                                               predicted_labels,
                                               average='weighted'),
                        4))
    print('Recall:', np.round(
                        metrics.recall_score(true_labels, 
                                               predicted_labels,
                                               average='weighted'),
                        4))
    print('F1 Score:', np.round(
                        metrics.f1_score(true_labels, 
                                               predicted_labels,
                                               average='weighted'),
                        4))
                        

def train_predict_model(classifier, 
                        train_features, train_labels, 
                        test_features, test_labels):
    # build model    
    classifier.fit(train_features, train_labels)
    # predict using model
    predictions = classifier.predict(test_features) 
    return predictions    


def display_confusion_matrix(true_labels, predicted_labels, classes=[1,0]):
    
    total_classes = len(classes)
    level_labels = [total_classes*[0], list(range(total_classes))]

    cm = metrics.confusion_matrix(y_true=true_labels, y_pred=predicted_labels, 
                                  labels=classes)
    cm_frame = pd.DataFrame(data=cm, 
                            columns=pd.MultiIndex(levels=[['Predicted:'], classes], 
                                                  labels=level_labels), 
                            index=pd.MultiIndex(levels=[['Actual:'], classes], 
                                                labels=level_labels)) 
    print(cm_frame) 
    
def display_classification_report(true_labels, predicted_labels, classes=[1,0]):

    report = metrics.classification_report(y_true=true_labels, 
                                           y_pred=predicted_labels, 
                                           labels=classes) 
    print(report)
    
    
    
def display_model_performance_metrics(true_labels, predicted_labels, classes=[1,0]):
    print('Model Performance metrics:')
    print('-'*30)
    get_metrics(true_labels=true_labels, predicted_labels=predicted_labels)
    print('\nModel Classification report:')
    print('-'*30)
    display_classification_report(true_labels=true_labels, predicted_labels=predicted_labels, 
                                  classes=classes)
    print('\nPrediction Confusion Matrix:')
    print('-'*30)
    display_confusion_matrix(true_labels=true_labels, predicted_labels=predicted_labels, 
                             classes=classes)


def plot_model_decision_surface(clf, train_features, train_labels,
                                plot_step=0.02, cmap=plt.cm.RdYlBu,
                                markers=None, alphas=None, colors=None):
    
    if train_features.shape[1] != 2:
        raise ValueError("X_train should have exactly 2 columnns!")
    
    x_min, x_max = train_features[:, 0].min() - plot_step, train_features[:, 0].max() + plot_step
    y_min, y_max = train_features[:, 1].min() - plot_step, train_features[:, 1].max() + plot_step
    xx, yy = np.meshgrid(np.arange(x_min, x_max, plot_step),
                         np.arange(y_min, y_max, plot_step))

    clf_est = clone(clf)
    clf_est.fit(train_features,train_labels)
    if hasattr(clf_est, 'predict_proba'):
        Z = clf_est.predict_proba(np.c_[xx.ravel(), yy.ravel()])[:,1]
    else:
        Z = clf_est.predict(np.c_[xx.ravel(), yy.ravel()])    
    Z = Z.reshape(xx.shape)
    cs = plt.contourf(xx, yy, Z, cmap=cmap)
    
    le = LabelEncoder()
    y_enc = le.fit_transform(train_labels)
    n_classes = len(le.classes_)
    plot_colors = ''.join(colors) if colors else [None] * n_classes
    label_names = le.classes_
    markers = markers if markers else [None] * n_classes
    alphas = alphas if alphas else [None] * n_classes
    for i, color in zip(range(n_classes), plot_colors):
        idx = np.where(y_enc == i)
        plt.scatter(train_features[idx, 0], train_features[idx, 1], c=color,
                    label=label_names[i], cmap=cmap, edgecolors='black', 
                    marker=markers[i], alpha=alphas[i])
    plt.legend()
    plt.show()


def plot_model_roc_curve(clf, features, true_labels, label_encoder=None, class_names=None):
    
    ## Compute ROC curve and ROC area for each class
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    if hasattr(clf, 'classes_'):
        class_labels = clf.classes_
    elif label_encoder:
        class_labels = label_encoder.classes_
    elif class_names:
        class_labels = class_names
    else:
        raise ValueError('Unable to derive prediction classes, please specify class_names!')
    n_classes = len(class_labels)
    y_test = label_binarize(true_labels, classes=class_labels)
    if n_classes == 2:
        if hasattr(clf, 'predict_proba'):
            prob = clf.predict_proba(features)
            y_score = prob[:, prob.shape[1]-1] 
        elif hasattr(clf, 'decision_function'):
            prob = clf.decision_function(features)
            y_score = prob[:, prob.shape[1]-1]
        else:
            raise AttributeError("Estimator doesn't have a probability or confidence scoring system!")
        
        fpr, tpr, _ = roc_curve(y_test, y_score)      
        roc_auc = auc(fpr, tpr)
        plt.plot(fpr, tpr, label='ROC curve (area = {0:0.2f})'
                                 ''.format(roc_auc),
                 linewidth=2.5)
        
    elif n_classes > 2:
        if hasattr(clf, 'predict_proba'):
            y_score = clf.predict_proba(features)
        elif hasattr(clf, 'decision_function'):
            y_score = clf.decision_function(features)
        else:
            raise AttributeError("Estimator doesn't have a probability or confidence scoring system!")

        for i in range(n_classes):
            fpr[i], tpr[i], _ = roc_curve(y_test[:, i], y_score[:, i])
            roc_auc[i] = auc(fpr[i], tpr[i])

        ## Compute micro-average ROC curve and ROC area
        fpr["micro"], tpr["micro"], _ = roc_curve(y_test.ravel(), y_score.ravel())
        roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

        ## Compute macro-average ROC curve and ROC area
        # First aggregate all false positive rates
        all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))
        # Then interpolate all ROC curves at this points
        mean_tpr = np.zeros_like(all_fpr)
        for i in range(n_classes):
            mean_tpr += interp(all_fpr, fpr[i], tpr[i])
        # Finally average it and compute AUC
        mean_tpr /= n_classes
        fpr["macro"] = all_fpr
        tpr["macro"] = mean_tpr
        roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

        ## Plot ROC curves
        plt.figure(figsize=(6, 4))
        plt.plot(fpr["micro"], tpr["micro"],
                 label='micro-average ROC curve (area = {0:0.2f})'
                       ''.format(roc_auc["micro"]), linewidth=3)

        plt.plot(fpr["macro"], tpr["macro"],
                 label='macro-average ROC curve (area = {0:0.2f})'
                       ''.format(roc_auc["macro"]), linewidth=3)

        for i, label in enumerate(class_labels):
            plt.plot(fpr[i], tpr[i], label='ROC curve of class {0} (area = {1:0.2f})'
                                           ''.format(label, roc_auc[i]), 
                     linewidth=2, linestyle=':')
    else:
        raise ValueError('Number of classes should be atleast 2 or more')
        
    plt.plot([0, 1], [0, 1], 'k--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver Operating Characteristic (ROC) Curve')
    plt.legend(loc="lower right")
    plt.show()


import os
import numpy as np
import pandas as pd
import tifffile as tfi
from skimage.util import img_as_ubyte
import string
import random
import skimage
from skimage.draw import disk


from AIPyS.classification.bayes.Baysian_training import Baysian_training
from AIPyS.supportFunctions.AIPS_file_display import Compsite_display
from AIPyS.supportFunctions.AIPS_functions import deployRatioCalc,call_bin,keepObject
from AIPyS.supportFunctions.GranularityFunc import openingOperation,resize,imageToRGB


class BayesDeploy(Baysian_training):
    '''
    on the fly cell call function for activating cells

    Parameters
    ----------
    file: str
        single channel target image
    path: str
    kernel_size: int
    trace_a: int
    trace_b: int
    thold: int
        probability threshold for calling cells
    pathOut: str
    clean: int
        remove object bellow the selected area size
    saveMerge: boolean

    Returns
    -------
    binary mask for activating the called cell
    '''
    def __init__(self,kernelGran,intercept,slope,*args, **kwargs):
        self.kernelGran = kernelGran
        self.intercept = intercept
        self.slope = slope
        super().__init__(*args, **kwargs)
        self.BayesianGranularityDeploy(saveMerge=True)
    
    def id_generator(self,size=6, chars=string.ascii_uppercase + string.digits):
        return ''.join(random.choice(chars) for _ in range(size))
    
    def sdCalc(self,image,mean_value):
        squared_diff_sum = 0
        num_elements = 0
        for row in image:
            for element in row:
                squared_diff_sum += (element - mean_value) ** 2
                num_elements += 1
        return (squared_diff_sum / num_elements)**0.5
    
    
    def check_singleImage(self):
        if isinstance(self.img, list):
            assert "should be a path to single image"

    def classify(self,n):
        mu = self.intercept + self.slope * n
        prob = 1 / (1 + np.exp(-mu))
        return prob
    
    def granularityDeploy(self,table,image,mask):
        img_blank = np.zeros_like(image,np.uint8)
        table["predict"] = 0.0
        table_out =  {col: [] for col in table.columns}
        active_label = []
        imgBound = (0,np.shape(img_blank)[0],np.shape(img_blank)[1])
        for idx, (index, row) in enumerate(table.iterrows()):
            img_label = row['label'] # first cell analysis 
            try:
                stack_img, stack_mask = self.stackObjects_cellpose_ebimage_parametrs_method(image,mask,table,img_label)
            except:
                continue
            if stack_img is None:
                continue
            intensity =  np.mean(stack_img)
            #image intensity and sd before opening:
            # # video gen
            imageDecy = openingOperation(kernel = self.kernelGran, image = stack_img)
            ratio = np.mean(imageDecy)/intensity
            prob = self.classify(ratio)
            row["predict"] = prob
            if prob < self.thold:
                # saves table with low pvalue for better analysis
                for col in table.columns:
                    table_out[col].append(row[col])
            else:
                active_label.append(img_label)
                x, y = row["centroid-0"], row["centroid-1"]
                r, c = disk((int(x), int(y)), 10)
                if np.any(np.in1d(imgBound, r)) or np.any(np.in1d(imgBound, c)):
                    for col in table.columns:
                        table_out[col].append(row[col])
                else:
                    img_blank[r, c] = 255
                    for col in table.columns:
                        table_out[col].append(row[col])
        return img_blank, pd.DataFrame(table_out),active_label
      
    def counts_files_check(self):
        filepath = os.path.join(self.outPath,'cell_count.txt')
        if not(os.path.exists(filepath)):
            with open(filepath,'w') as f:
                f.write(str(0))
        filepath = os.path.join(self.outPath,'count.txt')
        if not(os.path.exists(filepath)):
            with open(filepath,'w') as f:
                f.write(str(0))  
        filepath = os.path.join(self.outPath,'active_cell_count.txt')
        if not(os.path.exists(filepath)):
            with open(filepath,'w') as f:
                f.write(str(0))      
            
            
    def BayesianGranularityDeploy(self,saveMerge):
        self.check_singleImage()
        self.counts_files_check()
        pathOut = self.outPath
        image = self.img # must be single image
        mask, table = self.cellpose_segmentation(image_input = image)
        # intensity after decay
        if len(table) < 5:
            with open(os.path.join(pathOut, 'cell_count.txt'), 'r') as f:
                prev_number = f.readlines()
            new_value = int(prev_number[0]) + len(table)
            with open(os.path.join(pathOut, 'cell_count.txt'), 'w') as f:
                f.write(str(new_value))
            with open(os.path.join(pathOut, 'count.txt'), 'w') as f:
                f.write(str(len(table)))
        else:
            binary, table_sel, active_label =  self.granularityDeploy(table,image,mask)
            img_gs = img_as_ubyte(binary)
            if saveMerge:
                table_sel['predict'] = np.round(table_sel.predict.values, 2)
                #maskKeep = keepObject(table = table_sel,mask = mask)
                compsiteImage = Compsite_display(input_image=image, mask_roi=binary)
                LabeldImage = compsiteImage.display_image_label(table=table_sel, font_select="arial.ttf", font_size=14, label_draw='predict', intensity=1)
                LabeldImage.save(os.path.join(pathOut, self.id_generator() + '.png'))
            with open(os.path.join(pathOut, 'active_cell_count.txt'), 'r') as f:
                prev_number_active = f.readlines()
            new_value = int(prev_number_active[0]) + len(active_label)
            with open(os.path.join(pathOut, 'active_cell_count.txt'), 'w') as f:
                f.write(str(new_value))
            if os.path.exists(os.path.join(pathOut, 'binary.jpg')):
                os.remove(os.path.join(pathOut, 'binary.jpg'))
            skimage.io.imsave(os.path.join(pathOut, 'binary.jpg'), img_gs)
            with open(os.path.join(pathOut, 'cell_count.txt'), 'r') as f:
                prev_number = f.readlines()
            new_value = int(prev_number[0]) + len(table_sel)
            with open(os.path.join(pathOut, 'cell_count.txt'), 'w') as f:
                f.write(str(new_value))
            with open(os.path.join(pathOut, 'count.txt'), 'w') as f:
                f.write(str(len(table_sel)))
            print("done!")









# import tensorflow as tf
# from tensorflow.keras.applications.vgg16 import VGG16
# from tensorflow.keras.preprocessing import image
# from tensorflow.keras.applications.vgg16 import preprocess_input, decode_predictions
import numpy as np
from IPython.display import clear_output

# import glob
# import numpy as np
import os
import shutil
# import glob
# from sklearn.model_selection import train_test_split

np.random.seed(42)
# import matplotlib.pyplot as plt
# from tensorflow.keras.preprocessing.image import ImageDataGenerator, load_img, img_to_array, array_to_img

'''
Handle training data for DL binary classification
1) mix all the files in one location with proper label
2) 
'''

class classification_data_orgenizer:
    # organize images files and folder for binary classification
    # creates train, validation and test directories
    def __init__(self,report = None,path_input = None, path_origen = None,
                path_class_A = None, path_class_B = None,file_name_CA = None,file_name_CB = None,label_A = None,
                 label_B = None, file_extention= None):
        '''
        path_input - directory contain mixture of images for group a and b
        path_origen - training directory
        file_name_CA - list,str, file names group A
        file_name_CB - list,str, file names group A
        label_A - str ,example: 'norm'
        label_B - str , example: 'pheno'
        file_extention - str, example: 'png' or 'tif'
        '''
        self.report = report
        self.path_input = path_input
        self.path_origen = path_origen
        self.path_class_A = path_class_A
        self.path_class_B = path_class_B
        self.file_name_CA = file_name_CA
        self.file_name_CB = file_name_CB
        self.label_A = label_A
        self.label_B = label_B
        self.file_extention = file_extention


    def get_file_names_list(self):
        file_names = [file for file in os.listdir(self.path_input)
                    if file.endswith(self.file_extention)]
        self.file_name_CA = [file_name for file_name in file_names
                        if self.label_A in file_name]
        self.file_name_CB = [file_name for file_name in file_names
                        if self.label_B in file_name]
        return self.file_name_CA, self.file_name_CB

    def split_traning_set_and_copy(self):
        train_ratio = 0.80
        #test_ratio = 0.10
        validation_ratio = 0.10
        if len(self.file_name_CA) >= len(self.file_name_CB):
            size_train = int(len(self.file_name_CB)*train_ratio)
            size_val = int(len(self.file_name_CB)*validation_ratio)
        else:
            size_train = int(len(self.file_name_CA)*train_ratio)
            size_val = int(len(self.file_name_CA)*validation_ratio)
        A_train = np.random.choice(self.file_name_CA, size=size_train, replace=False)
        B_train = np.random.choice(self.file_name_CB, size=size_train, replace=False)
        A_val_test = list(set(self.file_name_CA) - set(A_train))
        B_val_test = list(set(self.file_name_CB) - set(B_train))
        A_val = np.random.choice(A_val_test, size=size_val, replace=False)
        B_val = np.random.choice(B_val_test, size=size_val, replace=False)
        A_test = list(set(A_val_test) - set(A_val))
        B_test = list(set(B_val_test) - set(B_val))
        statment_a = 'A set datasets: {}{}{}'.format(np.shape(A_train), np.shape(A_val),np.shape(A_test))
        statment_b = 'B set datasets: {}{}{}'.format(np.shape(B_train), np.shape(B_val),np.shape(B_test))
        # copy files to the folder
        os.chdir(self.path_origen)
        train_dir = 'training_data'
        val_dir = 'validation_data'
        test_dir = 'test_data'
        train_files = np.concatenate([A_train, B_train])
        validate_files = np.concatenate([A_val, B_val])
        test_files = np.concatenate([A_test, B_test])
        os.mkdir(train_dir) if not os.path.isdir(train_dir) else None
        os.mkdir(val_dir) if not os.path.isdir(val_dir) else None
        os.mkdir(test_dir) if not os.path.isdir(test_dir) else None
        
        train_dir = os.path.join(self.path_origen, 'training_data')
        val_dir = os.path.join(self.path_origen, 'validation_data')
        test_dir = os.path.join(self.path_origen, 'test_data')
        
        clear_output(wait=True) # remove all display 
        os.chdir(self.path_input)
        for i,fn in enumerate(train_files):
            shutil.copy(fn, train_dir)
            print('-' * i, end = '\r')
        print('{}'.format(fn[:5]))

        for i,fn in enumerate(validate_files):
            shutil.copy(fn, val_dir)
            print('-' * i, end = '\r')
        print('{}'.format(fn[:5]))

        for i,fn in enumerate(test_files):
            shutil.copy(fn, test_dir)
            print('-' * i, end = '\r')
        print('{}'.format(fn[:5]))
        return statment_a, statment_b, train_files, validate_files, test_files











        
   


'''
    Simulation of read counts using negative binomial distribution and preforming AIPS sorting and selecting cells.
    The effective sgRNA are selectd from a pool of sgRNA targeting genes. The number of True positive, FP rate and sample size are predetermined.
'''


import numpy as np
import os
import pandas as pd
import seaborn as sns
import matplotlib as plt
import random
import tqdm as tqdm

class Simulate():
    def __init__(self,df,lookupString = 'PEX',tpRatio  =10,n = 3, p=0.1):
        '''
        :param lookupString: str, substring of target Gene
        :param tpRatio, effective sgRNA number
        :param n, float, number of failures until the experiment is stopped
        :param n, float[0,1],success probability in each experiment
        '''
        self.df = df
        self.lookupString = lookupString
        self.tpRatio = tpRatio
        self.effectiveGuide = self.truePositiveTuple()
        self.n = n
        self.p = p
        self.dfSim = self.observePerRaw()

    def truePositiveTuple(self):
        '''
        :return tuple, sublist of effective sgRNA
        '''
        indexTarget = self.df.loc[self.df.gene.str.contains(self.lookupString)].index.to_list()
        self.df['activeSg'] = False
        indexPexActiveArray = self.df.loc[random.sample(indexTarget, self.tpRatio)].index.tolist()
        self.df.loc[indexPexActiveArray, 'activeSg'] = True
        # list of sgRNA which are true
        TruePositiveSGs = tuple(self.df.loc[indexPexActiveArray, 'sgID'].to_list())
        return TruePositiveSGs

    def observePerRaw(self):
        ''':param
        '''
        self.df['count_sim'] = np.random.negative_binomial(self.n, self.p, len(self.df))
        initSgRNA = self.df.sgID.values.tolist()
        initCount = self.df.count_sim.values.tolist()
        sgRNA = []
        Activity = []
        for sg,count in zip(initSgRNA,initCount):
            sgRNA += [sg]*count
            #
            if sg in self.effectiveGuide:
                Activity += [True]*count
            else:
                Activity += [False] * count
        Qoriginal = pd.DataFrame({'sgID': sgRNA, 'Active': Activity})
        table = Qoriginal.sample(frac=1, replace=False, random_state=1564743454, axis=0, ignore_index=True)
        return table

    def simulation(self,FalseLimits, ObservationNum):
        '''
        :param FalseLimits, tuple, precantage list of False Positive
        :param ObservationNum, tuple, mean and standard deviation
        '''
        Original = self.dfSim
        dfQ1 = {}
        dfQ2 = {}
        fpRate = [arr for arr in np.arange(FalseLimits[0], FalseLimits[1], FalseLimits[0])]
        progress = tqdm.tqdm()
        while True:
            progress.update()
            FOV = int(np.random.normal(ObservationNum[0],ObservationNum[1]))
            if FOV > len(self.dfSim):
                break
            dfTemp = self.dfSim.iloc[:FOV,:]
            # shorten the table by fov
            self.dfSim = self.dfSim.iloc[FOV+1:,:]
            idxTruePostive = dfTemp.index[dfTemp['Active']].tolist()
            if len(idxTruePostive) > 0:
                TruePositiveSGs = dfTemp.loc[idxTruePostive, 'sgID'].to_list()
                dfTemp = dfTemp.drop(idxTruePostive)
                for sg in TruePositiveSGs:
                    if sg in dfQ2.keys():
                        dfQ2[sg] += 1
                    else:
                        dfQ2[sg] = 1
            selFP = int(len(dfTemp.index.to_list()) * random.sample(fpRate,1)[0])
            if selFP > 0:
                TruePositiveSGs =  dfTemp['sgID'].sample(n = selFP).to_list()
                for sg in TruePositiveSGs:
                    dfTemp = dfTemp[dfTemp.sgID != sg]
                    if sg in dfQ2.keys():
                        dfQ2[sg] += 1
                    else:
                        dfQ2[sg] = 1
            sgRNAexclude =  dfTemp['sgID'].to_list()
            for sg in sgRNAexclude:
                if sg in dfQ1.keys():
                    dfQ1[sg] += 1
                else:
                    dfQ1[sg] = 1

        return Original,dfQ1,dfQ2



















import numpy as np
import os
import skimage.io
import glob
import pdb
import logging

class DataLoad:
    '''
    Takes a list of image names and returns a single image or a list of images. Note: Images must be 16-bit grayscale TIFF images, expecting glob. 
    Parameters
    ----------
    Image_name: str 
        glob module function
    outpath: str
    '''
    def __init__(self, Image_name = None, outPath = None):
        self.Image_name = Image_name
        self.outPath = outPath
        self.img = self.imageLoading()
    
    # def check_path(self,input):
    #     assert os.path.dirname(input), "The file_name object does not contain a path."
    
    def checkImageShape(self,imageInput):
        if len(imageInput.shape) > 2:
            raise ValueError("Images must be of single channel")
       
    
    def multiImageLoading(self):
        img = []
        for image in self.Image_name:
            img.append(self.singleimageLoading(imageName = image))
        return img
    
    def singleimageLoading(self,imageName):
        imgOut = skimage.io.imread(imageName)
        self.checkImageShape(imgOut)
        return imgOut
    
    def imageLoading(self):
        # self.check_path(self.Image_name)
        # self.check_path(self.outPath)
        if isinstance(self.Image_name, list): 
            return self.multiImageLoading()
        else:
            return self.singleimageLoading(imageName = self.Image_name)
            